import unittest
import os
pumadir = os.path.abspath('..')
import pumapy as puma

try:
    from TexGen.Core import *


    if os.path.exists(pumadir + "/python/TexGen/install/lib"):
        class TestWeaveIO(unittest.TestCase):

            def test_weave_io(self):
                # Create a textile
                Textile = CTextile()

                # Create a python list containing 4 yarns
                Yarns = [CYarn(), CYarn(), CYarn(), CYarn()]

                # Add nodes to the yarns to describe their paths
                Yarns[0].AddNode(CNode(XYZ(0, 0, 0)))
                Yarns[0].AddNode(CNode(XYZ(0.22, 0, 0.05)))
                Yarns[0].AddNode(CNode(XYZ(0.44, 0, 0)))

                Yarns[1].AddNode(CNode(XYZ(0, 0.22, 0.05)))
                Yarns[1].AddNode(CNode(XYZ(0.22, 0.22, 0)))
                Yarns[1].AddNode(CNode(XYZ(0.44, 0.22, 0.05)))

                Yarns[2].AddNode(CNode(XYZ(0, 0, 0.05)))
                Yarns[2].AddNode(CNode(XYZ(0, 0.22, 0)))
                Yarns[2].AddNode(CNode(XYZ(0, 0.44, 0.05)))

                Yarns[3].AddNode(CNode(XYZ(0.22, 0, 0)))
                Yarns[3].AddNode(CNode(XYZ(0.22, 0.22, 0.05)))
                Yarns[3].AddNode(CNode(XYZ(0.22, 0.44, 0)))

                # Loop over all the yarns in the list
                for Yarn in Yarns:
                    # Set the interpolation function
                    Yarn.AssignInterpolation(CInterpolationCubic())

                    # Assign a constant cross-section all along the yarn of elliptical shape
                    Yarn.AssignSection(CYarnSectionConstant(CSectionEllipse(0.18, 0.04)))

                    # Set the resolution of the surface mesh created
                    Yarn.SetResolution(20)

                    # Add repeat vectors to the yarn
                    Yarn.AddRepeat(XYZ(0.44, 0, 0))
                    Yarn.AddRepeat(XYZ(0, 0.44, 0))

                    # Add the yarn to our textile
                    Textile.AddYarn(Yarn)

                self.assertEqual(puma.export_weave_vtu(os.path.join(pumadir, "test", "out", "weavetest"), Textile,
                                                       CDomainPlanes(XYZ(0, 0, -0.02), XYZ(0.44, 0.44, 0.07)), 100),
                                 os.path.join(pumadir, "test", "out", "weavetest_100_100_21"))

                ws = puma.import_weave_vtu(os.path.join(pumadir, "test", "out", "weavetest_100_100_21.vtu"))
                self.assertEqual(max(ws.matrix.shape), 100)

except:
    puma.print_warning("Could not run test_weave_io because of failed TexGen import.")

if __name__ == '__main__':
    unittest.main()



from pumapy import Workspace
import pumapy.utilities.generic_checks as check
from pumapy.generation.sphere import get_sphere
import numpy as np
import sys


def generate_random_spheres(size, diameter, porosity, allow_intersect=True):
    """ Generation of triply periodic minimal surface material

    :param size: size of 3D domain (x,y,z)
    :type size: tuple(int, int, int)
    :param diameter: diameter of the random spheres in voxels
    :type diameter: float
    :param porosity: target porosity of the generated structure
    :type porosity: float
    :param allow_intersect: allow the spheres to intersect or not
    :type allow_intersect: bool
    :return: domain with random spheres with input diameter
    :rtype: Workspace

    :Example:
    >>> import pumapy as puma
    >>> ws_generated = puma.generate_random_spheres(size=(200,200,200), diameter=20, porosity=0.7) # Generating a workspace of randomly placed, intersecting spheres
    >>> puma.render_contour(ws_generated, cutoff=(128, 255))
    """

    generator = GeneratorSpheres(size, diameter, porosity, allow_intersect)

    generator.error_check()

    generator.log_input()
    generator.log_output()

    return generator.generate()


class GeneratorSpheres:
    def __init__(self, size, diameter, porosity, allow_intersect):
        self._workspace = Workspace()
        self._size = size
        self._diameter = diameter
        self._porosity = porosity
        self._allow_intersect = allow_intersect

        self._size_padded = (self._size[0]+4*diameter, self._size[1]+4*diameter, self._size[2]+4*diameter)
        self._start = 2 * diameter
        self._endX = size[0] + self._start
        self._endY = size[1] + self._start
        self._endZ = size[2] + self._start

        self._placed_sphere_list = [(-1e6, -1e6, -1e6)]

        # size of the created sphere block
        self._spheresize = int(self._diameter+4)
        if self._spheresize % 2 == 0:
            self._size_m = self._spheresize / 2
            self._size_p = self._spheresize / 2
        else:
            self._size_m = int(self._spheresize / 2.)+1
            self._size_p = int(self._spheresize / 2.)

    def generate(self):
        self._workspace.resize_new_matrix(self._size_padded)
        self._generate_spheres()

        return self._workspace

    def _generate_spheres(self):
        sphere = get_sphere(self._diameter)
        radius = self._diameter / 2.
        volume_per_sphere = 4./3. * np.pi * radius**3
        volume_all_spheres = self._size[0] * self._size[1] * self._size[2] * (1-self._porosity)

        approx_num_spheres = volume_all_spheres / volume_per_sphere / self._porosity
        print("Approximately " + str(approx_num_spheres) + " spheres to be generated")

        current_volume = 0
        approx_volume_left = volume_all_spheres - current_volume
        approx_spheres_left = approx_volume_left / volume_per_sphere

        current_porosity = 1
        localCount = 0
        totalCount = 0

        while current_porosity > self._porosity:
            x = np.random.randint(self._size[0] + 2 * self._diameter) + self._diameter
            y = np.random.randint(self._size[1] + 2 * self._diameter) + self._diameter
            z = np.random.randint(self._size[2] + 2 * self._diameter) + self._diameter
            xm = int(x - self._size_m)
            xp = int(x + self._size_p)
            ym = int(y - self._size_m)
            yp = int(y + self._size_p)
            zm = int(z - self._size_m)
            zp = int(z + self._size_p)

            if self._allow_intersect:
                self._workspace.matrix[xm:xp, ym:yp, zm:zp] += sphere
            else:
                intersection = False
                for location in self._placed_sphere_list:
                    if np.sqrt( (x-location[0])*(x-location[0]) +
                                (y-location[1])*(y-location[1]) +
                                (z-location[2])*(z-location[2]) ) < self._diameter:
                        intersection = True
                        break
                if intersection:
                    continue
                else:
                    self._workspace.matrix[xm:xp, ym:yp, zm:zp] += sphere
                    self._placed_sphere_list.append((x, y, z))

            localCount += 1
            totalCount += 1

            if localCount > 0.5 * approx_spheres_left:
                localCount = 0

                mask = self._workspace.matrix[self._start:self._endX, self._start:self._endY, self._start:self._endZ] >= 0
                mask_high = self._workspace.matrix[self._start:self._endX, self._start:self._endY, self._start:self._endZ] <= 127
                mask = mask * mask_high
                current_porosity = float(np.sum(mask)) / float(self._size[0] * self._size[1] * self._size[2])

                current_volume = self._size[0] * self._size[1] * self._size[2] * (1 - current_porosity)
                approx_volume_left = volume_all_spheres - current_volume
                approx_spheres_left = approx_volume_left / volume_per_sphere
                sys.stdout.write("\rSpheres Generated {}  Porosity = {} ".format(totalCount, current_porosity))

        self._workspace.matrix = self._workspace.matrix[self._start:self._endX, self._start:self._endY, self._start:self._endZ]
        self._workspace.matrix[self._workspace.matrix > 255] = 255
        self._workspace.matrix = self._workspace.matrix.astype(np.uint16)

    def log_input(self):
        self._workspace.log.log_section("Generating Random Spheres")
        self._workspace.log.log_value("Domain Size", self._size)
        self._workspace.log.log_value("Diameter", self._diameter)
        self._workspace.log.log_value("Porosity", self._porosity)
        self._workspace.log.write_log()

    def log_output(self):
        self._workspace.log.log_section("Finished Random Sphere Generation")
        self._workspace.log.write_log()

    def error_check(self):
        check.size_check(self._size)
        check.greater_than_exc(self._diameter, 0, "diameter")
        check.range_exc(self._porosity, (0, 1), "porosity")



import unittest
import numpy as np
import pumapy as puma
from pumapy.utilities.raycasting import RayCasting


class TestRayCasting(unittest.TestCase):

    def test_nophase(self):
        ws = puma.Workspace.from_shape_value((10, 10, 10), 1)
        simulation = RayCasting(ws, 45, np.array([[4, 4, 4]]), 0)
        try:
            simulation.error_check()
        except:
            print("Correctly caught exception.")
            return
        self.assertTrue(False)

    def test_non180divider(self):
        ws = puma.Workspace.from_shape_value((10, 10, 10), 1)
        simulation = RayCasting(ws, 70, np.array([[4, 4, 4]]), 1)
        try:
            simulation.error_check()
        except:
            print("Correctly caught exception.")
            return
        self.assertTrue(False)

    def test_numberofparticles(self):
        ws = puma.Workspace.from_shape_value((10, 10, 10), 1)
        simulation = RayCasting(ws, 45, np.array([[4, 4, 4]]), 1)
        self.assertFalse(simulation.error_check())
        simulation.generate_spherical_walkers()
        self.assertEqual(simulation.spherical_walkers.shape[0], 26)

    def test_distancetravelled1(self):
        ws = puma.Workspace.from_shape_value((9, 9, 9), 1)
        ws.voxel_length = 1
        simulation = RayCasting(ws, 90, np.array([[4, 4, 4]]), 1)
        self.assertFalse(simulation.error_check())
        simulation.generate_spherical_walkers()
        simulation.expand_sources()
        np.testing.assert_equal(np.sqrt(simulation.rays_distances[:, 0]**2 +
                                        simulation.rays_distances[:, 1]**2 +
                                        simulation.rays_distances[:, 2]**2), np.repeat(4.5, 6))

    def test_distancetravelled2(self):
        ws = puma.Workspace.from_shape_value((9, 9, 9), 1)
        ws.voxel_length = 1
        simulation = RayCasting(ws, 45, np.array([[4, 4, 4]]), 1)
        self.assertFalse(simulation.error_check())
        simulation.generate_spherical_walkers()
        simulation.expand_sources()
        np.testing.assert_almost_equal(np.sort(np.sqrt(simulation.rays_distances[:, 0]**2 +
                                                       simulation.rays_distances[:, 1]**2 +
                                                       simulation.rays_distances[:, 2]**2)),
                                       np.concatenate((np.repeat(4.5, 6), np.repeat(np.sqrt(2) * 4.5, 20))), 7)

    def test_distancetravelled_withcollision(self):
        ws = puma.Workspace.from_shape_value((9, 9, 9), 1)
        ws[4, 4, 0] = 0
        ws[4, 1, 4] = 0
        ws.voxel_length = 1
        source = np.array([[4, 4, 4]])
        simulation = RayCasting(ws, 90, source, 1)
        self.assertFalse(simulation.error_check())
        simulation.generate_spherical_walkers()
        simulation.expand_sources()
        np.testing.assert_almost_equal(np.sort(np.sqrt(simulation.rays_distances[:, 0]**2 +
                                                       simulation.rays_distances[:, 1]**2 +
                                                       simulation.rays_distances[:, 2]**2)),
                                       np.concatenate((np.array([2.5, 3.5]), np.repeat(4.5, 4))), 7)


if __name__ == '__main__':
    unittest.main()



from pumapy.utilities.workspace import Workspace
from pumapy.utilities.boundary_conditions import ConductivityBC
from pumapy.utilities.linear_solvers import PropertySolver
import numpy as np


class Conductivity(PropertySolver):

    def __init__(self, workspace, cond_map, direction, side_bc, prescribed_bc, tolerance, maxiter, solver_type, display_iter):
        allowed_solvers = ['direct', 'gmres', 'cg', 'bicgstab']
        super().__init__(workspace, solver_type, allowed_solvers, tolerance, maxiter, display_iter)

        self.cond_map = cond_map
        self.direction = direction
        self.side_bc = side_bc
        self.prescribed_bc = prescribed_bc

        self.keff = [-1., -1., -1.]
        self.solve_time = -1
        self.T = np.zeros([1, 1, 1])
        self.q = np.zeros([1, 1, 1, 3])

    def log_input(self):
        self.ws.log.log_section("Computing Conductivity")
        self.ws.log.log_line("Domain Size: " + str(self.ws.get_shape()))
        self.ws.log.log_line("Conductivity Map: ")
        for i in range(self.cond_map.get_size()):
            low, high, cond = self.cond_map.get_material(i)
            self.ws.log.log_line(
                "  - Material " + str(i) + "[" + str(low) + "," + str(high) + "," + str(cond) + "]")
            self.ws.log.log_line("Solver Type: " + str(self.solver_type))
        self.ws.log.log_line("Solver Tolerance: " + str(self.tolerance))
        self.ws.log.log_line("Max Iterations: " + str(self.maxiter))
        self.ws.log.write_log()

    def log_output(self):
        self.ws.log.log_section("Finished Conductivity Calculation")
        self.ws.log.log_line("Conductivity: " + "[" + str(self.keff) + "]")
        self.ws.log.log_line("Solver Time: " + str(self.solve_time))
        self.ws.log.write_log()

    def error_check(self):
        # ws checks
        if not isinstance(self.ws, Workspace):
            raise Exception("Workspace must be a puma.Workspace.")
        if self.ws.len_x() < 3 or self.ws.len_y() < 3 or self.ws.len_z() < 3:
            raise Exception("Workspace must be at least 3x3x3 for Conductivity solver.")

        # direction checks
        if self.direction == "x" or self.direction == "X":
            self.direction = "x"
        elif self.direction == "y" or self.direction == "Y":
            self.direction = "y"
        elif self.direction == "z" or self.direction == "Z":
            self.direction = "z"
        else:
            raise Exception("Invalid simulation direction.")

        # side_bc checks
        if self.side_bc == "periodic" or self.side_bc == "Periodic" or self.side_bc == "p":
            self.side_bc = "p"
        elif self.side_bc == "symmetric" or self.side_bc == "Symmetric" or self.side_bc == "s":
            self.side_bc = "s"
        elif self.side_bc == "dirichlet" or self.side_bc == "Dirichlet" or self.side_bc == "d":
            self.side_bc = "d"
        else:
            raise Exception("Invalid side boundary conditions.")

        # prescribed_bc checks
        if self.prescribed_bc is not None:
            if not isinstance(self.prescribed_bc, ConductivityBC):
                raise Exception("prescribed_bc must be a puma.ConductivityBC.")
            if self.prescribed_bc.dirichlet.shape != self.ws.matrix.shape:
                raise Exception("prescribed_bc must be of the same size as the domain.")

            # rotate it
            if self.direction == 'y':
                self.prescribed_bc.dirichlet = self.prescribed_bc.dirichlet.transpose((1, 0, 2))
            elif self.direction == 'z':
                self.prescribed_bc.dirichlet = self.prescribed_bc.dirichlet.transpose((2, 1, 0))
            if np.any((self.prescribed_bc[[0, -1]] == np.Inf)):
                raise Exception("prescribed_bc must be defined on the direction sides")



from pumapy.physicsmodels.anisotropic_conductivity_utils import pad_domain
from pumapy.physicsmodels.elasticity_utils import (fill_stress_matrices, flatten_Cmat, add_nondiag, divP,
                                                   find_unstable_vox)
from pumapy.physicsmodels.mpxa_matrices import fill_Ampsa, fill_Bmpsa, fill_Cmpsa, fill_Dmpsa, create_mpsa_indices
from pumapy.utilities.workspace import Workspace
from pumapy.utilities.boundary_conditions import ElasticityBC
from pumapy.utilities.linear_solvers import PropertySolver
from pumapy.utilities.timer import Timer
from scipy.sparse import csr_matrix, diags
import numpy as np
import sys


class Elasticity(PropertySolver):

    def __init__(self, workspace, elast_map, direction, side_bc, prescribed_bc, tolerance, maxiter, solver_type,
                 display_iter, print_matrices):
        allowed_solvers = ['direct', 'gmres', 'cg', 'bicgstab']
        super().__init__(workspace, solver_type, allowed_solvers, tolerance, maxiter, display_iter)

        self.elast_map = elast_map
        self.direction = direction
        self.side_bc = side_bc
        self.prescribed_bc = prescribed_bc
        self.print_matrices = print_matrices
        self.mat_elast = dict()
        self.need_to_orient = False  # changes if (E_axial, E_radial, nu_poissrat_12, nu_poissrat_23, G12) detected
        self.orient_pad = None

        self.Ceff = [-1., -1., -1.]
        self.solve_time = -1
        self.u = np.zeros([1, 1, 1])
        self.s = np.zeros([1, 1, 1, 3])
        self.t = np.zeros([1, 1, 1, 3])

    def compute(self):
        t = Timer()
        self.initialize()
        self.assemble_bvector()
        self.assemble_Amatrix()
        print("Time to assemble matrices: ", t.elapsed()); t.reset()
        super().solve()
        print("Time to solve: ", t.elapsed())
        self.compute_effective_coefficient()
        self.solve_time = t.elapsed()

    def initialize(self):
        print("Initializing and padding domains ... ", flush=True, end='')

        # Rotating domain to avoid cases and padding
        shape = [self.len_x + 2, self.len_y + 2, self.len_z + 2]
        reorder = [0, 1, 2]
        if self.direction == 'y' or self.direction == 'z':
            if self.direction == 'y':
                shape = [self.len_y + 2, self.len_z + 2, self.len_x + 2]
                reorder = [1, 2, 0]
                a1, b1, g1, a2, b2, g2, a3, b3, g3 = (0, 1, 0, 0, 0, 1, 1, 0, 0)
            else:
                shape = [self.len_z + 2, self.len_x + 2, self.len_y + 2]
                reorder = [2, 0, 1]
                a1, b1, g1, a2, b2, g2, a3, b3, g3 = (0, 0, 1, 1, 0, 0, 0, 1, 0)

            # Rotating matelast
            C = np.zeros((6, 6), dtype=float)
            R = np.array([[a1 ** 2, a2 ** 2, a3 ** 2, -2 * a2 * a3, -2 * a3 * a1, -2 * a1 * a2],
                          [b1 ** 2, b2 ** 2, b3 ** 2, -2 * b2 * b3, -2 * b3 * b1, -2 * b1 * b2],
                          [g1 ** 2, g2 ** 2, g3 ** 2, -2 * g2 * g3, -2 * g3 * g1, -2 * g1 * g2],
                          [-b1 * g1, -b2 * g2, -b3 * g3, b2 * g3 + b3 * g2, b1 * g3 + b3 * g1, b1 * g2 + b2 * g1],
                          [-g1 * a1, -g2 * a2, -g3 * a3, g2 * a3 + g3 * a2, g1 * a3 + g3 * a1, g1 * a2 + g2 * a1],
                          [-a1 * b1, -a2 * b2, -a3 * b3, a2 * b3 + a3 * b2, a1 * b3 + a3 * b1, a1 * b2 + a2 * b1]])
            for key, value in self.mat_elast.items():
                if len(value) == 21:
                    C[np.triu_indices(6)] = value
                    C[np.tril_indices(6, -1)] = np.tril(C.T, -1)[np.tril_indices(6, -1)]
                    C = R.T @ C @ R
                    self.mat_elast[key] = tuple(C[np.triu_indices(6)])

        self.ws_pad = np.zeros(shape, dtype=np.uint16)
        self.ws_pad[1:-1, 1:-1, 1:-1] = self.ws.matrix.transpose(reorder)

        if self.need_to_orient:
            self.orient_pad = np.zeros(shape + [3], dtype=float)
            self.orient_pad[1:-1, 1:-1, 1:-1, :] = self.ws.orientation.transpose(reorder + [3])[:, :, :, reorder]

        self.len_x, self.len_y, self.len_z = shape
        self.len_xy = self.len_x * self.len_y
        self.len_xyz = self.len_x * self.len_y * self.len_z

        if self.side_bc is not "f":
            # Padding domain, imposing symmetric or periodic BC on faces
            pad_domain(self.ws_pad, self.orient_pad, self.need_to_orient, self.len_x, self.len_y, self.len_z,
                       self.side_bc)

        # Segmenting padded domain
        for i in range(self.elast_map.get_size()):
            low, high, _ = self.elast_map.get_material(i)
            self.ws_pad[np.logical_and(self.ws_pad >= low, self.ws_pad <= high)] = low

        # Placing True on dirichlet boundaries to skip them
        self.dir_vox = np.zeros(shape + [3], dtype=bool)
        if self.direction is not None:
            self.dir_vox[[1, -2], 1:-1, 1:-1] = True
        if self.prescribed_bc is not None:
            self.dir_vox[1:-1, 1:-1, 1:-1][self.prescribed_bc.dirichlet != np.Inf] = True
        print("Done")

        # Initialize initial guess for iterative solver
        if self.solver_type != 'direct' and self.solver_type != 'spsolve':
            self.initial_guess = np.zeros((self.len_x, self.len_y, self.len_z, 3), dtype=float)
            for i in range(self.len_x - 1):
                self.initial_guess[i, :, :, 0] = i / (self.len_x - 2.)
            self.initial_guess = self.initial_guess.flatten('F')

    def assemble_bvector(self):
        print("Assembling b vector ... ", flush=True, end='')

        I, V = ([] for _ in range(2))

        if self.prescribed_bc is not None:
            for i in range(1, self.len_x - 1):
                for j in range(1, self.len_y - 1):
                    for k in range(1, self.len_z - 1):
                        if self.prescribed_bc[i - 1, j - 1, k - 1, 0] != np.Inf:
                            I.append(self.len_x * (self.len_y * k + j) + i)
                            V.append(self.prescribed_bc[i - 1, j - 1, k - 1, 0])  # ux
                        if self.prescribed_bc[i - 1, j - 1, k - 1, 1] != np.Inf:
                            I.append(self.len_xyz + self.len_x * (self.len_y * k + j) + i)
                            V.append(self.prescribed_bc[i - 1, j - 1, k - 1, 1])  # uy
                        if self.prescribed_bc[i - 1, j - 1, k - 1, 2] != np.Inf:
                            I.append(2 * self.len_xyz + self.len_x * (self.len_y * k + j) + i)
                            V.append(self.prescribed_bc[i - 1, j - 1, k - 1, 2])  # uz
        else:
            # Setting unit displacement
            i = self.len_x - 2
            for j in range(1, self.len_y - 1):
                for k in range(1, self.len_z - 1):
                    I.append(self.len_x * (self.len_y * k + j) + i)
                    V.append(1.)

        # Setting linear displacement on the boundaries if Dirichlet
        if self.side_bc == 'd' and self.direction is not None:
            x = np.linspace(0, 1, self.len_x - 2)
            for j in [1, self.len_y - 2]:
                for i in range(1, self.len_x - 1):
                    for k in range(1, self.len_z - 1):
                        I.append(self.len_x * (self.len_y * k + j) + i)
                        V.append(x[i - 1])
            for k in [1, self.len_z - 2]:
                for i in range(2, self.len_x - 2):
                    for j in range(2, self.len_y - 2):
                        I.append(self.len_x * (self.len_y * k + j) + i)
                        V.append(x[i - 1])

        self.bvec = csr_matrix((V, (I, np.zeros(len(I)))), shape=(3 * self.len_xyz, 1))

        if self.print_matrices[0]:
            self._print_b(self.print_matrices[0])
        print("Done")

    def assemble_Amatrix(self):
        print("Initializing large data structures ... ", flush=True, end='')
        I, J = np.zeros((2, 81 * 3 * self.len_xyz), dtype=np.uint32)
        V = np.zeros(81 * 3 * self.len_xyz, dtype=float)
        counter = 0  # counter to keep record of the index in Amat
        I_dirvox = []
        self.__initialize_MPSA()
        j_indices = np.zeros((81 * 3 * (self.len_y - 2) * (self.len_z - 2)), dtype=np.uint32)
        values = np.zeros((81 * 3 * (self.len_y - 2) * (self.len_z - 2)), dtype=float)
        self.dir_vox = self.dir_vox.astype(np.uint8)
        print("Done")

        # Iterating through interior
        for i in range(1, self.len_x - 1):
            self.__compute_Cmat(2, i + 1)  # Computing third layer of Cmat
            self.__compute_transmissibility(1, i)  # Computing second layer of E

            # If all surrounding IV are unstable (i.e. partly or all gaseous), then put middle CV as Dirichlet
            find_unstable_vox(i, self.len_y, self.len_z, self.dir_vox, self.unstable)

            # Creating j indices and divergence values for slice
            j_indices.fill(-1)
            values.fill(np.NaN)
            divP(i, self.len_x, self.len_y, self.len_z, self.dir_vox, j_indices, values, self.Emat)

            # Creating i indices for slice
            i_indices, i_dirvox = self.__creating_indices(i)
            if i_indices.size > 0:
                I[counter:counter + i_indices.size] = i_indices
            I_dirvox.extend(i_dirvox)

            if j_indices[j_indices != -1].size > 0:
                J[counter:counter + i_indices.size] = j_indices[~np.isnan(values)]
                V[counter:counter + i_indices.size] = values[~np.isnan(values)]
                counter += i_indices.size

            # Passing second layer to first
            self.Emat[0] = self.Emat[1]
            self.unstable[0] = self.unstable[1]
            self.Cmat[:2] = self.Cmat[1:]
            sys.stdout.write("\rAssembling A matrix ... {:.1f}% ".format(i / (self.len_x - 2) * 100))

        # Clear unnecessary variables before creating A
        del self.Emat, self.Cf, self.Cmat, self.mpsa36x36, self.unstable
        del self.dir_vox, i_indices, i_dirvox, i, j_indices, values

        # Adding all dirichlet voxels
        I[counter:counter + len(I_dirvox)] = I_dirvox
        J[counter:counter + len(I_dirvox)] = I_dirvox
        V[counter:counter + len(I_dirvox)] = 1
        counter += len(I_dirvox)
        del I_dirvox

        # Add diagonal 1s for exterior voxels
        diag_1s = np.ones_like(self.ws_pad, dtype=int)
        diag_1s[1:-1, 1:-1, 1:-1] = 0  # interior to 0
        ind = np.array(np.where(diag_1s > 0))  # indices of contour
        diag_1s = self.len_x * (self.len_y * ind[2] + ind[1]) + ind[0]
        diag_1s = np.hstack((diag_1s, self.len_xyz + diag_1s, 2 * self.len_xyz + diag_1s))
        diag_1s = diag_1s.astype(np.uint32)
        del ind
        I[counter:counter + diag_1s.size] = diag_1s
        J[counter:counter + diag_1s.size] = diag_1s
        V[counter:counter + diag_1s.size] = 1
        counter += diag_1s.size

        # Add non-diagonal 1s for exterior voxels
        if self.side_bc is not "d" and self.side_bc is not "f":
            I[counter:counter + diag_1s.size] = diag_1s
            nondiag1s = np.ones_like(diag_1s, dtype=np.int8)
            add_nondiag(diag_1s, nondiag1s, self.len_x, self.len_y, self.len_z, self.side_bc)
            J[counter:counter + diag_1s.size] = diag_1s  # CAREFUL: diag_1s reused for nondiag to optimize memory
            if self.side_bc == "s":
                V[counter:counter + diag_1s.size] = nondiag1s
            else:
                V[counter:counter + diag_1s.size] = -1
            del nondiag1s
        del diag_1s, counter

        # Assemble sparse A matrix
        self.Amat = csr_matrix((V, (I, J)), shape=(3 * self.len_xyz, 3 * self.len_xyz))

        # Simple preconditioner
        diag = self.Amat.diagonal()
        if np.any(diag == 0):
            self.M = None  # identity matrix if singularity has happened in MPSA
        else:
            self.M = diags(1. / self.Amat.diagonal(), 0).tocsr()

        if self.print_matrices[2]:
            self._print_A(self.print_matrices[2])
        print("Done")

    def compute_effective_coefficient(self):
        # reshaping solution
        self.u = self.x.reshape([self.len_x, self.len_y, self.len_z, 3], order='F')
        del self.x

        # Mirroring boundaries for flux computation
        if self.direction is not None:
            self.u[0] = self.u[1]
            self.u[-1] = self.u[-2]
            if self.side_bc == "d" or self.side_bc == "f":
                self.u[:, 0] = self.u[:, 1]
                self.u[:, -1] = self.u[:, -2]
                self.u[:, :, 0] = self.u[:, :, 1]
                self.u[:, :, -1] = self.u[:, :, -2]
        if self.print_matrices[3]:
            show_u(self.u, self.print_matrices[3])
        print(" ... Done")

        self.__compute_stresses()

        if self.direction is not None:
            # Accumulating and volume averaging stresses
            stresses = [np.sum(self.s[:, :, :, i]) / ((self.len_x - 2) * (self.len_y - 2) * (self.len_z - 2)) for i in
                        range(3)]
            stresses += [np.sum(self.t[:, :, :, i]) / ((self.len_x - 2) * (self.len_y - 2) * (self.len_z - 2)) for i in
                         range(3)]
            self.Ceff = [stresses[i] * (self.len_x - 2) * self.ws.voxel_length for i in range(6)]

            # Rotating output back
            if self.direction == 'y':
                self.u = self.u.transpose(2, 0, 1, 3)[:, :, :, [2, 0, 1]]
                self.s = self.s.transpose(2, 0, 1, 3)[:, :, :, [2, 0, 1]]
                self.t = self.t.transpose(2, 0, 1, 3)[:, :, :, [2, 0, 1]]
                self.Ceff = [self.Ceff[2], self.Ceff[0], self.Ceff[1], self.Ceff[5], self.Ceff[3], self.Ceff[4]]
            elif self.direction == 'z':
                self.u = self.u.transpose(1, 2, 0, 3)[:, :, :, [1, 2, 0]]
                self.s = self.s.transpose(1, 2, 0, 3)[:, :, :, [1, 2, 0]]
                self.t = self.t.transpose(1, 2, 0, 3)[:, :, :, [1, 2, 0]]
                self.Ceff = [self.Ceff[1], self.Ceff[2], self.Ceff[0], self.Ceff[4], self.Ceff[5], self.Ceff[3]]

            d = {'x': 'first', 'y': 'second', 'z': 'third'}
            print(f'\nEffective elasticity tensor ({d[self.direction]} row): \n{self.Ceff}')

    def index_at(self, index, size):
        if self.side_bc == "p":
            if index == 0:
                return size - 1
            elif index == size:
                return 1
        else:
            if index == 0:
                return 1
            elif index == size:
                return size - 1
        return index

    def __compute_Cmat(self, i, i_cv):
        # Reset layer of Cmat
        self.Cmat[i].fill(0)

        # Assigning elasticities throughout domain
        for key, value in self.mat_elast.items():
            mask = self.ws_pad[i_cv] == key
            if len(value) == 21:
                self.Cmat[i, mask] = value

            else:  # local elasticity orientation
                E1, E2, v12, v23, G12 = value
                v21 = v12 * E2 / E1
                delta = 1 - 2 * v12 * v21 - v23 * v23 - 2 * v21 * v12 * v23
                C_tmp = np.array([[((1 - v23 * v23) * E1) / delta, (v21 * (1 + v23) * E1) / delta,
                                   (v21 * (1 + v23) * E1) / delta, 0, 0, 0],
                                  [(v21 * (1 + v23) * E1) / delta, ((1 - v12 * v21) * E2) / delta,
                                   ((v23 + v21 * v12) * E2) / delta, 0, 0, 0],
                                  [(v21 * (1 + v23) * E1) / delta, ((v23 + v21 * v12) * E2) / delta,
                                   ((1 - v21 * v12) * E2) / delta, 0, 0, 0],
                                  [0, 0, 0, ((1 - v23 - 2 * v21 * v12) * E2) / delta, 0, 0],
                                  [0, 0, 0, 0, 2 * G12, 0],
                                  [0, 0, 0, 0, 0, 2 * G12]])
                size = np.sum(mask)
                C_init = np.repeat(C_tmp[:, :, np.newaxis], size, axis=2)

                # Rotation matrix
                theta = np.arctan2(self.orient_pad[i_cv, mask, 1], self.orient_pad[i_cv, mask, 0])
                a21 = -np.sin(theta)
                a22 = np.cos(theta)
                beta = np.arcsin(self.orient_pad[i_cv, mask, 2])
                a13 = np.sin(beta)
                a33 = np.cos(beta)
                a11 = a22 * a33
                a12 = - a21 * a33
                a31 = - a22 * a13
                a32 = a21 * a13
                a23 = np.zeros(size)
                R = np.array([[a11 ** 2, a12 ** 2, a13 ** 2, a12 * a13, a11 * a13, a11 * a12],
                              [a21 ** 2, a22 ** 2, a23, a23, a23, a21 * a22],
                              [a31 ** 2, a32 ** 2, a33 ** 2, a32 * a33, a33 * a31, a31 * a32],
                              [2 * a21 * a31, 2 * a32 * a22, a23, a22 * a33,
                               a21 * a33, a21 * a32 + a22 * a31],
                              [2 * a11 * a31, 2 * a12 * a32, 2 * a13 * a33, a32 * a13 + a33 * a12,
                               a11 * a33 + a13 * a31, a31 * a12 + a32 * a11],
                              [2 * a11 * a21, 2 * a12 * a22, a23, a13 * a22,
                               a13 * a21, a11 * a22 + a12 * a21]])

                C_final = R.transpose((2, 1, 0)) @ C_init.transpose((2, 0, 1)) @ R.transpose((2, 0, 1))
                ind = np.triu_indices(6)
                self.Cmat[i, mask] = C_final[:, ind[0], ind[1]]

    def __compute_transmissibility(self, i, i_cv):
        # Reset layers
        self.Emat[i].fill(0)
        self.unstable[i].fill(0)
        self.Cf.fill(0)
        flatten_Cmat(i, self.len_y, self.len_z, self.Cmat[i:i + 2], self.Cf)

        # C becomes singular sometimes when there are air voxels
        self.mpsa36x36.fill(0)
        self.mpsa36x36[:, :, self.Cind[0], self.Cind[1]] = fill_Cmpsa(self.Cf)
        det = np.linalg.det(self.mpsa36x36)
        if np.min(det) < 1e-10:
            self.unstable[i, det < 1e-10] = True

        # Computing transmissibility matrix as: (A @ (Cinv @ D) + B)/8
        if not np.all(self.unstable[i]):
            self.Emat[i, :, :, self.Dind[0], self.Dind[1]] = fill_Dmpsa(self.Cf)
            self.Emat[i, self.unstable[i]] = 0

            self.mpsa36x36[~self.unstable[i]] = np.linalg.inv(self.mpsa36x36[~self.unstable[i]])  # Cinv

            self.Emat[i, ~self.unstable[i]] = (self.mpsa36x36[~self.unstable[i]] @
                                               self.Emat[i, ~self.unstable[i]])  # (Cinv @ D)

            self.mpsa36x36.fill(0)
            self.mpsa36x36[:, :, self.Aind[0], self.Aind[1]] = fill_Ampsa(self.Cf)
            self.Emat[i, ~self.unstable[i]] = (self.mpsa36x36[~self.unstable[i]] @
                                               self.Emat[i, ~self.unstable[i]])  # A @ (Cinv @ D)

            self.Emat[i, ~self.unstable[i]] += fill_Bmpsa(self.Cf)[~self.unstable[i]]  # + B
            self.Emat[i, ~self.unstable[i]] /= 8

        if self.print_matrices[1]:
            self._print_E(i, i_cv, self.print_matrices[1])

    def __initialize_MPSA(self):
        # Initialize matrix slice of conductivities
        self.Cmat = np.zeros((3, self.len_y, self.len_z, 21), dtype=float)  # per CV
        self.__compute_Cmat(0, 0)  # Computing first layer of Kmat
        self.__compute_Cmat(1, 1)  # Computing second layer of Kmat

        # Initialize MPSA variables
        self.Cf = np.zeros((168, self.len_y - 1, self.len_z - 1), dtype=float)  # per IV
        self.Emat = np.zeros((2, self.len_y - 1, self.len_z - 1, 36, 24), dtype=float)
        self.unstable = np.zeros((2, self.len_y - 1, self.len_z - 1), dtype=bool)
        self.mpsa36x36 = np.zeros((self.len_y - 1, self.len_z - 1, 36, 36), dtype=float)  # A, C
        self.Aind, self.Cind, self.Dind = create_mpsa_indices()
        self.__compute_transmissibility(0, 0)  # Computing first layer of E

    def __creating_indices(self, i):
        # Finding all indices for slice
        i_indices = np.ones_like(self.ws_pad[i], dtype=np.uint32)
        i_indices[[0, -1], :] = 0
        i_indices[:, [0, -1]] = 0
        i_indices = np.where(i_indices > 0)
        i_indices = self.len_x * (self.len_y * i_indices[1] + i_indices[0]) + np.full(i_indices[0].size, i)
        i_indices = np.hstack((i_indices, self.len_xyz + i_indices, 2 * self.len_xyz + i_indices))

        # Removing dirichlet voxels
        i_dirvox = np.where(self.dir_vox[i, :, :, 0])
        i_dirvox = self.len_x * (self.len_y * i_dirvox[1] + i_dirvox[0]) + np.full(i_dirvox[0].size, i)
        i_dirvox1 = np.where(self.dir_vox[i, :, :, 1])
        i_dirvox1 = self.len_x * (self.len_y * i_dirvox1[1] + i_dirvox1[0]) + np.full(i_dirvox1[0].size, i)
        i_dirvox = np.hstack((i_dirvox, self.len_xyz + i_dirvox1))
        i_dirvox1 = np.where(self.dir_vox[i, :, :, 2])
        i_dirvox1 = self.len_x * (self.len_y * i_dirvox1[1] + i_dirvox1[0]) + np.full(i_dirvox1[0].size, i)
        i_dirvox = np.hstack((i_dirvox, 2 * self.len_xyz + i_dirvox1))
        i_indices = i_indices[~np.in1d(i_indices, i_dirvox)]

        # Duplicating the voxel indices where divergence happens
        i_indices = np.repeat(i_indices, 81)
        return i_indices, i_dirvox  # returning dirichlet voxel indices

    def __compute_stresses(self):
        # Initialize required data structures
        self.s, self.t = np.zeros((2, self.len_x - 2, self.len_y - 2, self.len_z - 2, 3))
        self.__initialize_MPSA()
        u_sw, u_se, u_nw, u_ne, u_tsw, u_tse, u_tnw, u_tne = np.zeros((8, self.len_y - 2, self.len_z - 2, 24))  # per CV
        E_sw, E_se, E_nw, E_ne, E_tsw, E_tse, E_tnw, E_tne = np.zeros((8, self.len_y - 2, self.len_z - 2, 36, 24))

        # Computing first layer of E
        self.__compute_transmissibility(0, 0)

        # Iterating through interior
        for i in range(1, self.len_x - 1):
            self.__compute_Cmat(2, i + 1)  # Computing third layer of Cmat
            self.__compute_transmissibility(1, i)  # Computing second layer of E

            # filling eight IVs
            fill_stress_matrices(i, self.len_x, self.len_y, self.len_z, self.u[i - 1:i + 2], self.Emat,
                                 E_sw, E_se, E_nw, E_ne, E_tsw, E_tse, E_tnw, E_tne,
                                 u_sw, u_se, u_nw, u_ne, u_tsw, u_tse, u_tnw, u_tne)

            # Computing stresses by computing E @ u
            s_sw = np.squeeze(E_sw @ u_sw[:, :, :, np.newaxis])
            s_se = np.squeeze(E_se @ u_se[:, :, :, np.newaxis])
            s_nw = np.squeeze(E_nw @ u_nw[:, :, :, np.newaxis])
            s_ne = np.squeeze(E_ne @ u_ne[:, :, :, np.newaxis])
            s_tsw = np.squeeze(E_tsw @ u_tsw[:, :, :, np.newaxis])
            s_tse = np.squeeze(E_tse @ u_tse[:, :, :, np.newaxis])
            s_tnw = np.squeeze(E_tnw @ u_tnw[:, :, :, np.newaxis])
            s_tne = np.squeeze(E_tne @ u_tne[:, :, :, np.newaxis])

            # Summing fluxes CV-wise
            self.s[i - 1, :, :, 0] = (s_se[:, :, 9] + s_ne[:, :, 6] + s_tse[:, :, 3] + s_tne[:, :, 0] +
                                      s_sw[:, :, 9] + s_nw[:, :, 6] + s_tsw[:, :, 3] + s_tnw[:, :, 0])  # sigma_x
            self.s[i - 1, :, :, 1] = (s_nw[:, :, 21] + s_ne[:, :, 18] + s_tnw[:, :, 15] + s_tne[:, :, 12] +
                                      s_sw[:, :, 21] + s_se[:, :, 18] + s_tsw[:, :, 15] + s_tse[:, :, 12])  # sigma_y
            self.s[i - 1, :, :, 2] = (s_tsw[:, :, 33] + s_tse[:, :, 30] + s_tnw[:, :, 27] + s_tne[:, :, 24] +
                                      s_sw[:, :, 33] + s_se[:, :, 30] + s_nw[:, :, 27] + s_ne[:, :, 24])  # sigma_z

            self.t[i - 1, :, :, 0] = (s_tsw[:, :, 34] + s_tse[:, :, 31] + s_tnw[:, :, 28] + s_tne[:, :, 25] +
                                      s_sw[:, :, 34] + s_se[:, :, 31] + s_nw[:, :, 28] + s_ne[:, :, 25] +
                                      s_nw[:, :, 23] + s_ne[:, :, 20] + s_tnw[:, :, 17] + s_tne[:, :, 14] +
                                      s_sw[:, :, 23] + s_se[:, :, 20] + s_tsw[:, :, 17] + s_tse[:, :, 14])  # tau_yz
            self.t[i - 1, :, :, 1] = (s_tsw[:, :, 35] + s_tse[:, :, 32] + s_tnw[:, :, 29] + s_tne[:, :, 26] +
                                      s_sw[:, :, 35] + s_se[:, :, 32] + s_nw[:, :, 29] + s_ne[:, :, 26] +
                                      s_se[:, :, 10] + s_ne[:, :, 7] + s_tse[:, :, 4] + s_tne[:, :, 1] +
                                      s_sw[:, :, 10] + s_nw[:, :, 7] + s_tsw[:, :, 4] + s_tnw[:, :, 1])  # tau_xz
            self.t[i - 1, :, :, 2] = (s_nw[:, :, 22] + s_ne[:, :, 19] + s_tnw[:, :, 16] + s_tne[:, :, 13] +
                                      s_sw[:, :, 22] + s_se[:, :, 19] + s_tsw[:, :, 16] + s_tse[:, :, 13] +
                                      s_se[:, :, 11] + s_ne[:, :, 8] + s_tse[:, :, 5] + s_tne[:, :, 2] +
                                      s_sw[:, :, 11] + s_nw[:, :, 8] + s_tsw[:, :, 5] + s_tnw[:, :, 2])  # tau_xy

            # Passing second layer to first
            self.Emat[0] = self.Emat[1]
            self.Cmat[:2] = self.Cmat[1:]
            sys.stdout.write("\rComputing stresses ... {:.1f}% ".format(i / (self.len_x - 2) * 100))
        del self.Emat, self.Cf, self.Cmat, self.mpsa36x36, self.unstable

        self.s /= 8 * self.ws.voxel_length
        self.t /= 16 * self.ws.voxel_length

        if self.print_matrices[4]:
            show_s(self.s, self.t, self.print_matrices[4])

        # Extract only interior displacement, ignoring exterior used as bc
        self.u = self.u[1:-1, 1:-1, 1:-1]
        print("Done")

    def log_input(self):
        self.ws.log.log_section("Computing Elasticity")
        self.ws.log.log_line("Simulation direction: " + str(self.direction))
        self.ws.log.log_line("Domain Size: " + str(self.ws.get_shape()))
        self.ws.log.log_line("Elasticity Map: ")
        for i in range(self.elast_map.get_size()):
            low, high, cond = self.elast_map.get_material(i)
            self.ws.log.log_line("  - Material " + str(i) + "[" + str(low) + "," + str(high) + "," + str(cond) + "]")
        self.ws.log.log_line("Solver Type: " + str(self.solver_type))
        self.ws.log.log_line("Solver Tolerance: " + str(self.tolerance))
        self.ws.log.log_line("Max Iterations: " + str(self.maxiter))
        self.ws.log.write_log()

    def log_output(self):
        self.ws.log.log_section("Finished Elasticity Calculation")
        self.ws.log.log_line("Elasticity: " + "[" + str(self.Ceff) + "]")
        self.ws.log.log_line("Solver Time: " + str(self.solve_time))
        self.ws.log.write_log()

    def error_check(self):
        # ws checks
        if not isinstance(self.ws, Workspace):
            raise Exception("Workspace must be a puma.Workspace.")
        if self.ws.len_x() < 3 or self.ws.len_y() < 3 or self.ws.len_z() < 3:
            raise Exception("Workspace must be at least 3x3x3 for Elasticity MPSA method.")

        # elast_map checks
        ws_tmp_tocheck = self.ws.matrix.copy()
        for i in range(self.elast_map.get_size()):
            low, high, C = self.elast_map.get_material(i)
            self.mat_elast[low] = C
            if len(C) == 5:
                self.need_to_orient = True
                if self.ws.orientation.shape[:3] != self.ws.matrix.shape or \
                        not 0.9 < np.min(np.linalg.norm(self.ws.orientation[np.logical_and(self.ws.matrix >= low,
                                                                                           self.ws.matrix <= high)],
                                                        axis=1)) < 1.1:
                    raise Exception("The Workspace needs an orientation in order to align the elasticity.")

            # segmenting tmp domain to check if all values covered by mat_elast
            ws_tmp_tocheck[np.logical_and(self.ws.matrix >= low, self.ws.matrix <= high)] = low

        unique_matrixvalues = np.unique(ws_tmp_tocheck)
        if 0 in unique_matrixvalues and 0 not in self.mat_elast.keys():
            self.elast_map.add_isotropic_material((0, 0), 0., 0.)
            _, _, self.mat_elast[0] = self.elast_map.get_material(self.elast_map.get_size() - 1)
        if (unique_matrixvalues.size != len(self.mat_elast.keys()) or
                np.all(np.sort(list(self.mat_elast.keys())).astype(np.uint16) != unique_matrixvalues)):
            raise Exception("All values in workspace must be included in ElasticityMap.")

        # side_bc checks
        if self.side_bc == "periodic" or self.side_bc == "Periodic" or self.side_bc == "p":
            self.side_bc = "p"
        elif self.side_bc == "symmetric" or self.side_bc == "Symmetric" or self.side_bc == "s":
            self.side_bc = "s"
        elif self.side_bc == "dirichlet" or self.side_bc == "Dirichlet" or self.side_bc == "d":
            self.side_bc = "d"
        elif self.side_bc == "free" or self.side_bc == "Free" or self.side_bc == "f":
            self.side_bc = "f"
        else:
            raise Exception("Invalid side boundary conditions.")

        # direction checks
        if self.direction is not None:
            if self.direction == "x" or self.direction == "X":
                self.direction = "x"
            elif self.direction == "y" or self.direction == "Y":
                self.direction = "y"
            elif self.direction == "z" or self.direction == "Z":
                self.direction = "z"
            else:
                raise Exception("Invalid simulation direction.")

        # print_matrices checks
        if type(self.print_matrices) is not tuple or len(self.print_matrices) != 5:
            raise Exception("Print_matrices must be a tuple with 5 booleans.")

        # prescribed_bc checks
        if self.prescribed_bc is not None:
            if not isinstance(self.prescribed_bc, ElasticityBC):
                raise Exception("prescribed_bc must be a puma.ElasticityBC.")
            if self.prescribed_bc.dirichlet.shape[:3] != self.ws.matrix.shape:
                raise Exception("prescribed_bc must be of the same size as the domain.")

            # rotate it
            if self.direction == 'y':
                self.prescribed_bc = self.prescribed_bc.dirichlet.transpose((1, 0, 2, 3))
            elif self.direction == 'z':
                self.prescribed_bc = self.prescribed_bc.dirichlet.transpose((2, 1, 0, 3))

            if self.direction is not None:
                if np.any((self.prescribed_bc[[0, -1]] == np.Inf)):
                    raise Exception("prescribed_bc must be defined on the direction sides")
        else:
            if self.direction is None:
                raise Exception("prescribed_bc must be defined for compute_stress_analysis")

    # Printing functions of system matrices
    def _print_E(self, i, i_cv, dec=4):
        np.set_printoptions(precision=dec)
        np.set_printoptions(linewidth=10000)
        for j in range(self.len_y - 1):
            for k in range(self.len_z - 1):
                print()
                print("index {},{},{}".format(i_cv, j, k))
                print(self.Emat[i, j, k])

    def _print_A(self, dec=4):
        np.set_printoptions(linewidth=10000)
        np.set_printoptions(threshold=sys.maxsize)
        np.set_printoptions(formatter={'float': lambda x: "{:.{}f}".format(x, dec).rstrip('0').rstrip('.')})
        print()
        print("A matrix:")
        print(self.Amat.toarray())

    def _print_b(self, dec=1):
        print()
        print("b vector:")
        print("  o---> y")
        print("  |")
        print("x v")
        for k in range(self.len_z):
            for i in range(self.len_x):
                for j in range(self.len_y):
                    print('({:.{}f}, {:.{}f}, {:.{}f})'.format(self.bvec[self.len_x * (self.len_y * k + j) + i, 0], dec,
                                                               self.bvec[self.len_xyz + self.len_x * (
                                                                       self.len_y * k + j) + i, 0], dec,
                                                               self.bvec[2 * self.len_xyz + self.len_x * (
                                                                       self.len_y * k + j) + i, 0], dec), end=' ')
                print()
            print()


def show_u(u, dec=4):
    print()
    print("3D Displacement:")
    print("  o---> y")
    print("  |")
    print("x v")
    for k in range(u.shape[2]):
        for i in range(u.shape[0]):
            for j in range(u.shape[1]):
                print('({:.{}f}, {:.{}f}, {:.{}f})'.format(u[i, j, k, 0], dec,
                                                           u[i, j, k, 1], dec,
                                                           u[i, j, k, 2], dec), end=' ')
            print()
        print()


def show_s(s, t, dec=4):
    print()
    print("3D Stress:")
    print("  o---> y")
    print("  |")
    print("x v")
    for k in range(s.shape[2]):
        for i in range(s.shape[0]):
            for j in range(s.shape[1]):
                print('({:.{}f}, {:.{}f}, {:.{}f}, {:.{}f}, {:.{}f}, {:.{}f})'.format(s[i, j, k, 0], dec,
                                                                                      s[i, j, k, 1], dec,
                                                                                      s[i, j, k, 2], dec,
                                                                                      t[i, j, k, 0], dec,
                                                                                      t[i, j, k, 1], dec,
                                                                                      t[i, j, k, 2], dec), end=' ')
            print()
        print()



from pumapy.utilities.logger import print_warning
from pumapy.utilities.timer import Timer
from pumapy.utilities.boundary_conditions import Isotropic_periodicBC, Isotropic_symmetricBC
from pumapy.physicsmodels.conductivity_parent import Conductivity
from pumapy.physicsmodels.isotropic_conductivity_utils import setup_matrices_cy, compute_flux
from scipy.sparse import csr_matrix, diags
import numpy as np


class IsotropicConductivity(Conductivity):

    def __init__(self, workspace, cond_map, direction, side_bc, prescribed_bc, tolerance, maxiter, solver_type,
                 display_iter):
        super().__init__(workspace, cond_map, direction, side_bc, prescribed_bc, tolerance, maxiter, solver_type,
                         display_iter)
        self._bc_func = None
        self.cond = np.zeros([1, 1, 1])

        if self.direction == 'x':
            self._matrix = workspace.matrix
        elif self.direction == 'y':
            self._matrix = workspace.matrix.transpose(1, 0, 2)
        elif self.direction == 'z':
            self._matrix = workspace.matrix.transpose(2, 1, 0)

        self.len_x, self.len_y, self.len_z = self._matrix.shape
        self.len_xy = self.len_x * self.len_y
        self.len_xyz = self.len_xy * self.len_z

        self.bc_check = 0
        if self.side_bc == "p" or self.side_bc == "periodic":
            self._bc_func = Isotropic_periodicBC(self.len_x, self.len_y, self.len_z)
        elif self.side_bc == "s" or self.side_bc == "symmetric":
            self._bc_func = Isotropic_symmetricBC(self.len_x, self.len_y, self.len_z)
        else:
            print_warning("Side Boundary Condition not recognized. Defaulting to symmetric")
            self._bc_func = Isotropic_symmetricBC(self.len_x, self.len_y, self.len_z)

        self.n_iter = 0
        self.Amat = None
        self.M = None
        self.bvec = np.zeros(1)
        self.initial_guess = np.zeros(1)
        self._kf = np.zeros(1)

    def compute(self):
        t = Timer()
        self.initialize()
        self.assemble_bvector()
        self.assemble_Amatrix()
        print("Time to assemble matrices: ", t.elapsed()); t.reset()
        super().solve()
        print("Time to solve: ", t.elapsed())
        self.compute_effective_coefficient()
        self.solve_time = t.elapsed()

    def initialize(self):
        print("Creating conductivity matrix ... ", end='')
        self.cond = np.zeros(self._matrix.shape, dtype=float)
        for i in range(self.cond_map.get_size()):
            low, high, k = self.cond_map.get_material(i)
            mask_low = self._matrix >= low
            mask_high = self._matrix <= high
            mask_low = mask_low * mask_high
            self.cond[mask_low] = k
        print("Done")

        print("Initializing temperature field ... ", end='')
        self.T = np.zeros([self.len_x, self.len_y, self.len_z])
        for i in range(self.len_x):
            self.T[i, :, :] = i / (self.len_x - 1.)
        self.initial_guess = self.T.flatten('F')
        print("Done")

    def assemble_bvector(self):

        print("Setting up b matrix ... ", end='')
        bsq = np.zeros([self.len_x, self.len_y, self.len_z])
        if self.prescribed_bc is not None:
            for i in range(self.len_x):
                for j in range(self.len_y):
                    for k in range(self.len_z):
                        if self.prescribed_bc[i, j, k] != np.Inf:
                            bsq[i, j, k] = self.prescribed_bc[i, j, k]

            self.prescribed_bc = self.prescribed_bc.dirichlet  # because of cython, cannot pass object
            self.bc_check = 1
        else:
            self.bc_check = 0
            self.prescribed_bc = np.full(self._matrix.shape, np.Inf, dtype=float)
            bsq[-1, :, :] = 1

        # check for zero conductivity
        for i in range(self.cond_map.get_size()):
            low, high, k = self.cond_map.get_material(i)
            if k == 0:
                self.bc_check = 1
                self.prescribed_bc[(self._matrix >= low) * (self._matrix <= high)] = 0

        self.bvec = bsq.flatten('F')
        print("Done")

    def assemble_Amatrix(self):
        self._kf = self.cond.flatten('F')
        self._row, self._col, self._data = setup_matrices_cy(self._kf, self.len_x, self.len_y, self.len_z,
                                                             self.bc_check, self.prescribed_bc)
        n_elem = self.len_xyz
        self.Amat = csr_matrix((self._data, (self._row, self._col)), shape=(n_elem, n_elem))
        print("Done")

        print("Setting up preconditioner ...", end ='')
        diag = self.Amat.diagonal()
        diag[diag==0] = 1
        diag[:] = 1./diag[:]
        self.M = diags(diag, 0).tocsr()
        print("Done")

    def compute_effective_coefficient(self):
        # reshaping solution
        self.T = self.x.reshape([self.len_x, self.len_y, self.len_z], order='F')
        del self.x

        if self.direction == 'y':
            self.T = self.T.transpose(1, 0, 2)
            self.cond = self.cond.transpose(1, 0, 2)
            self.len_x = self.T.shape[0]
            self.len_y = self.T.shape[1]
            self.len_z = self.T.shape[2]
        elif self.direction == 'z':
            self.T = self.T.transpose(2, 1, 0)
            self.cond = self.cond.transpose(2, 1, 0)
            self.len_x = self.T.shape[0]
            self.len_y = self.T.shape[1]
            self.len_z = self.T.shape[2]

        print("Computing flux from converged temperature field ... ", end='')
        flux_x, flux_y, flux_z, self.q = compute_flux(self.T, self.cond, self.len_x, self.len_y, self.len_z)
        print("Done")

        self.keff[0] = flux_x * (self.len_x - 1)
        self.keff[1] = flux_y * (self.len_y - 1)
        self.keff[2] = flux_z * (self.len_z - 1)

        d = {'x': 'first', 'y': 'second', 'z': 'third'}
        print(f'\nEffective conductivity tensor ({d[self.direction]} row): \n{self.keff}')

        # making the flux have the correct spacial units
        self.q /= self.ws.voxel_length



class MaterialPropertyMap:
    def __init__(self):
        self._lowCutoffs = []
        self._highCutoffs = []
        self._properties = []

    def error_check(self, cutoff, conductivity):
        if not isinstance(cutoff, tuple) or len(cutoff) != 2:
            raise Exception("Cutoff has to be a tuple(int,int) specifying the low and high cutoffs")

        for i in range(len(self._lowCutoffs)):
            if (self._highCutoffs[i] >= cutoff[0] >= self._lowCutoffs[i]) \
                    or (self._highCutoffs[i] >= cutoff[1] >= self._lowCutoffs[i]):
                raise Exception("Invalid material range. The range overlaps an existing material range")

        check = False
        if isinstance(conductivity, tuple):
            if any(i < 0 for i in conductivity):
                check = True
        else:
            if conductivity < 0:
                check = True
        if check:
            raise Exception("Invalid conductivity. Must be positive")
        return conductivity

    def _append_inputs(self, cutoff, conductivity):
        self._lowCutoffs.append(cutoff[0])
        self._highCutoffs.append(cutoff[1])
        self._properties.append(conductivity)

    def get_size(self):
        return len(self._lowCutoffs)

    def get_material(self, i):
        if i >= len(self._lowCutoffs):
            raise Exception("Invalid index. Maximum size: " + str(self.get_size()))
        if i < 0:
            raise Exception("Invalid index. Must be >= 0")

        return self._lowCutoffs[i], self._highCutoffs[i], self._properties[i]

    def show(self):
        print("Material conductivity as [low-high cutoffs] = conductivity:")
        for i in range(len(self._lowCutoffs)):
            print('[{} - {}] = {}'.format(self._lowCutoffs[i], self._highCutoffs[i], self._properties[i]))


class IsotropicConductivityMap(MaterialPropertyMap):
    def __init__(self):
        super().__init__()

    def add_material(self, cutoff, conductivity):
        conductivity = self.error_check(cutoff, conductivity)
        if isinstance(conductivity, bool):
            return
        self._append_inputs(cutoff, conductivity)


class AnisotropicConductivityMap(MaterialPropertyMap):
    def __init__(self):
        super().__init__()

    def add_material(self, cutoff, kxx, kyy, kzz, kxy, kxz, kyz):
        conductivity = (kxx, kyy, kzz, kxy, kxz, kyz)
        conductivity = self.error_check(cutoff, conductivity)
        if isinstance(conductivity, bool):
            return
        self._append_inputs(cutoff, conductivity)

    def add_isotropic_material(self, cutoff, k):
        conductivity = (k, k, k, 0., 0., 0.)
        conductivity = self.error_check(cutoff, conductivity)
        if isinstance(conductivity, bool):
            return
        self._append_inputs(cutoff, conductivity)

    def add_orthotropic_material(self, cutoff, kxx, kyy, kzz):
        conductivity = (kxx, kyy, kzz, 0., 0., 0.)
        conductivity = self.error_check(cutoff, conductivity)
        if isinstance(conductivity, bool):
            return
        self._append_inputs(cutoff, conductivity)

    def add_material_to_orient(self, cutoff, k_axial, k_radial):
        conductivity = (k_axial, k_radial)
        conductivity = self.error_check(cutoff, conductivity)
        if isinstance(conductivity, bool):
            return
        self._append_inputs(cutoff, conductivity)


class ElasticityMap(MaterialPropertyMap):
    def __init__(self):
        super().__init__()

    def add_material(self, cutoff, C11, C12, C13, C14, C15, C16, C22, C23, C24, C25, C26,
                     C33, C34, C35, C36, C44, C45, C46, C55, C56, C66):

        elasticity = (C11, C12, C13, C14, C15, C16,
                      C22, C23, C24, C25, C26,
                      C33, C34, C35, C36,
                      C44, C45, C46,
                      C55, C56,
                      C66)
        elasticity = self.error_check(cutoff, elasticity)
        if isinstance(elasticity, bool):
            return
        self._append_inputs(cutoff, elasticity)

    def add_isotropic_material(self, cutoff, E_youngmod, nu_poissrat):
        Lambda = (nu_poissrat * E_youngmod) / ((1 + nu_poissrat) * (1 - 2 * nu_poissrat))
        mu = E_youngmod / (2 * (1 + nu_poissrat))

        elasticity = (Lambda + 2 * mu, Lambda, Lambda, 0., 0., 0.,
                      Lambda + 2 * mu, Lambda,     0., 0., 0.,
                      Lambda + 2 * mu,     0.,     0., 0.,
                      2 * mu,              0.,     0.,
                      2 * mu,              0.,
                      2 * mu)
        elasticity = self.error_check(cutoff, elasticity)
        if isinstance(elasticity, bool):
            return
        self._append_inputs(cutoff, elasticity)

    def add_material_to_orient(self, cutoff, E_axial, E_radial, nu_poissrat_12, nu_poissrat_23, G12):
        elasticity = (E_axial, E_radial, nu_poissrat_12, nu_poissrat_23, G12)
        elasticity = self.error_check(cutoff, elasticity)
        if isinstance(elasticity, bool):
            return
        self._append_inputs(cutoff, elasticity)

    def show(self):
        print("Material elasticity as [low-high cutoffs] = elasticity:")
        for i in range(len(self._lowCutoffs)):
            print('[{} - {}] = '.format(self._lowCutoffs[i], self._highCutoffs[i], self._properties[i]))
            if len(self._properties[i]) == 5:
                print('{}'.format(self._properties[i]))
            else:
                first_elast, last_elast = (0, 6)
                for i2 in range(6):
                    print('{}'.format(self._properties[i][first_elast:last_elast]))
                    first_elast = last_elast
                    last_elast += 5 - i2



import unittest
import pumapy as puma


class TestOrientationST(unittest.TestCase):

    def test_artfib(self):
        ws1 = puma.import_vti("testdata/artifib.vtk")
        puma.compute_orientation_st(ws1, 0.7, 1.1, (128, 255))
        ws2 = puma.import_vti("testdata/artifib_orient.vtk")
        error, mean, std = puma.compute_angular_differences(ws1.matrix, ws1.orientation, ws2.orientation, (128, 255))
        self.assertTrue(mean < 5 and std < 10)

    def test_artfib_edt(self):
        ws1 = puma.import_vti("testdata/artifib.vtk")
        puma.compute_orientation_st(ws1, 0.7, 1.1, (128, 255), edt=True)
        ws2 = puma.import_vti("testdata/artifib_orient.vtk")
        error, mean, std = puma.compute_angular_differences(ws1.matrix, ws1.orientation, ws2.orientation, (128, 255))
        print(mean, std)
        self.assertTrue(mean < 6 and std < 10)


if __name__ == '__main__':
    unittest.main()



from pumapy.physicsmodels.mpsa_elasticity import Elasticity
from pumapy.utilities.property_maps import ElasticityMap


def compute_elasticity(workspace, elast_map, direction, side_bc='p', prescribed_bc=None, tolerance=1e-4,
                       maxiter=10000, solver_type='bicgstab', display_iter=True, print_matrices=(0, 0, 0, 0, 0)):
    """ Compute the thermal conductivity (N.B. 0 material ID in workspace refers to air unless otherwise specified)

    :param workspace: domain
    :type workspace: Workspace
    :param elast_map: local elasticity of the constituents
    :type elast_map: ElasticityMap
    :param direction: direction for solve ('x','y', 'z', 'yz', 'xz', 'xy')
    :type direction: string
    :param side_bc: side boundary conditions can be symmetric ('s'), periodic ('p'), dirichlet ('d') or free ('f')
    :type side_bc: string, optional
    :param prescribed_bc: 3D array holding dirichlet BC
    :type prescribed_bc: ElasticityBC, optional
    :param tolerance: tolerance for iterative solver
    :type: tolerance: float, optional
    :param maxiter: maximum Iterations for solver
    :type maxiter: int, optional
    :param solver_type: solver type, options: 'bicgstab', 'cg', 'gmres', 'direct'
    :type solver_type: string, optional
    :param display_iter: display iterations and residual
    :type display_iter: bool, optional
    :param print_matrices: corresponding to b, E, A, u, s decimal places. If 0, they are not printed
    :type print_matrices: tuple(5 ints), optional
    :return: elasticity, displacement field, direct stresses, shear stresses
    :rtype: tuple(tuple(6 floats), ndarray, ndarray, ndarray)

    :Example:
    >>> import pumapy as puma
    >>> export_name = 'halfmat'
    >>> X = 20
    >>> Y = 20
    >>> Z = 20
    >>> ws = puma.Workspace.from_shape_value((X, Y, Z), 1)
    >>> ws[int(X / 2):] = 2
    >>> elast_map = puma.ElasticityMap()
    >>> elast_map.add_isotropic_material((1, 1), 200, 0.3)
    >>> elast_map.add_isotropic_material((2, 2), 400, 0.1)
    >>> C, u, s, t = puma.compute_elasticity(ws, elast_map, direction='x', side_bc='f', solver_type="direct")
    >>> print(C)
    """
    if isinstance(elast_map, ElasticityMap):
        solver = Elasticity(workspace, elast_map, direction, side_bc, prescribed_bc, tolerance, maxiter,
                            solver_type, display_iter, print_matrices)
    else:
        raise Exception("elast_map has to be an ElasticityMap")

    solver.error_check()

    solver.log_input()
    solver.compute()
    solver.log_output()
    return solver.Ceff, solver.u, solver.s, solver.t


def compute_stress_analysis(workspace, elast_map, prescribed_bc=None, side_bc='p', tolerance=1e-4,
                            maxiter=10000, solver_type='bicgstab', display_iter=True, print_matrices=(0, 0, 0, 0, 0)):
    """ Compute the thermal conductivity (N.B. 0 material ID in workspace refers to air unless otherwise specified)

    :param workspace: domain
    :type workspace: Workspace
    :param elast_map: local elasticity of the constituents
    :type elast_map: ElasticityMap
    :param prescribed_bc: 3D array holding dirichlet BC
    :type prescribed_bc: ElasticityBC, optional
    :param side_bc: side boundary conditions can be symmetric ('s'), periodic ('p'), dirichlet ('d') or free ('f')
    :type side_bc: string, optional
    :param tolerance: tolerance for iterative solver
    :type tolerance: float, optional
    :param maxiter: maximum Iterations for solver
    :type maxiter: int, optional
    :param solver_type: solver type, options: 'bicgstab', 'cg', 'gmres', 'direct'
    :type solver_type: string, optional
    :param display_iter: display iterations and residual
    :type display_iter: bool, optional
    :param print_matrices: corresponding to b, E, A, u, s decimal places. If 0, they are not printed
    :type print_matrices: tuple(5 ints), optional
    :return: displacement field, direct stresses, shear stresses 'yz', 'xz', 'xy'
    :rtype: tuple(ndarray, ndarray, ndarray)
    """
    if isinstance(elast_map, ElasticityMap):
        solver = Elasticity(workspace, elast_map, None, side_bc, prescribed_bc, tolerance, maxiter,
                            solver_type, display_iter, print_matrices)
    else:
        raise Exception("elast_map has to be an ElasticityMap")

    solver.error_check()

    solver.log_input()
    solver.compute()
    solver.log_output()
    return solver.u, solver.s, solver.t



from pumapy.utilities.raycasting import RayCasting
from scipy.optimize import curve_fit
from matplotlib import pyplot as plt
import numpy as np
from pumapy import Workspace
from pumapy.utilities.logger import print_warning


def compute_radiation(workspace, solid_cutoff, sources_number, degree_accuracy, void_phase=0, boundary_behavior=1,
                      bin_density=10000, exportparticles_filepathname='', export_pathname=None):
    """ Compute the radiative thermal conductivity through ray tracing
    (N.B. 0 material ID in workspace refers to gas phases unless otherwise specified)

    :param workspace: domain
    :type workspace: Workspace
    :param solid_cutoff: specify the solid phase
    :type solid_cutoff: tuple(int, int)
    :param sources_number: number of light sources spread randomly in the void space (i.e. 0)
    :type sources_number: int
    :param degree_accuracy: angle difference between rays emitted in degrees (has to be an exact divider of 180°)
    :type degree_accuracy: int
    :param void_phase: ID of the void phase, defaulted as 0
    :type void_phase: int, optional
    :param boundary_behavior: how to treat particles exiting the domain: 0=kill, 1=periodic (default)
    :type boundary_behavior: int, optional
    :param bin_density: number of bins used to create histogram of ray distances
    :type bin_density: int, optional
    :param exportparticles_filepathname: path and name of the particle files exported as VTK points
    :type exportparticles_filepathname: string, optional
    :param export_pathname: path to save curve plot of ray distance distribution
    :type export_pathname: str, optional
    :return: extinction coefficient (beta), its standard deviation and of ray distances
    :rtype: tuple(float, float, ndarray)
    """
    solver = Radiation(workspace, solid_cutoff, sources_number, degree_accuracy, void_phase, boundary_behavior, bin_density,
                       exportparticles_filepathname, export_pathname)

    solver.error_check()

    solver.log_input()
    rays_distances = solver.compute()
    solver.log_output()
    return solver.beta, solver.beta_std, rays_distances


class Radiation:

    def __init__(self, workspace, cutoff, sources_number, degree_accuracy, void_phase, boundary_behavior, bin_density,
                 rayexport_filepathname, export_plot):
        self.workspace = workspace
        self.cutoff = cutoff
        self.sources_number = sources_number
        self.degree_accuracy = degree_accuracy
        self.void_phase = void_phase
        self.boundary_behavior = boundary_behavior
        self.bin_density = bin_density
        self.rayexport_filepathname = rayexport_filepathname
        self.export_pathname = export_plot
        self.X = None
        self.Y = None
        self.Z = None
        self.beta = [None, None, None]
        self.beta_std = [None, None, None]

    def compute(self):

        simulation = RayCasting(self.workspace, self.degree_accuracy, self.generate_sources(), self.void_phase,
                                self.boundary_behavior, self.rayexport_filepathname)

        simulation.error_check()

        simulation.generate_spherical_walkers()
        simulation.expand_sources()
        if self.export_pathname is not None:
            np.save(self.export_pathname, simulation.rays_distances)

        self.beta, self.beta_std = compute_extinction_coefficients(self.workspace, simulation.rays_distances,
                                                                   self.sources_number, self.degree_accuracy,
                                                                   self.bin_density, self.export_pathname)
        return simulation.rays_distances

    def generate_sources(self):
        # randomly choosing the source locations
        void_voxels_number = np.count_nonzero(self.workspace.matrix == self.void_phase)
        source_locations = np.array(np.where(self.workspace.matrix == self.void_phase)).transpose()
        if void_voxels_number <= self.sources_number:
            print_warning("Too many sources, limiting it to the number of void voxels: {}".format(void_voxels_number))
        else:
            choices = np.random.choice(source_locations.shape[0], size=self.sources_number, replace=False)
            source_locations = source_locations[choices]
        return source_locations

    def error_check(self):
        if not isinstance(self.workspace, Workspace):
            raise Exception("Workspace must be a puma.Workspace.")
        else:
            self.X = self.workspace.matrix.shape[0]
            self.Y = self.workspace.matrix.shape[1]
            self.Z = self.workspace.matrix.shape[2]
            self.workspace.binarize_range(self.cutoff)

        if np.count_nonzero(self.workspace.matrix == self.void_phase) == 0:
            raise Exception("No valid voxels detected (i.e. ID={}), cannot run radiation ray tracing.".format(self.void_phase))

    def log_input(self):
        self.workspace.log.log_section("Computing Radiation")
        self.workspace.log.log_line("Domain Size: " + str(self.workspace.get_shape()))
        self.workspace.log.log_line("Sources: " + str(self.sources_number))
        self.workspace.log.log_line("Degree accuracy: " + str(self.degree_accuracy))
        self.workspace.log.write_log()

    def log_output(self):
        self.workspace.log.log_section("Finished Radiation Simulation")
        self.workspace.log.log_line("Extinction coefficient beta: [" +
                                    str(self.beta[0]) + ' +/- ' + str(self.beta_std[0]) + ', ' +
                                    str(self.beta[1]) + ' +/- ' + str(self.beta_std[1]) + ', ' +
                                    str(self.beta[2]) + ' +/- ' + str(self.beta_std[2]) + ']')
        self.workspace.log.write_log()


def compute_extinction_coefficients(ws, rays_distances, sources_number, degree_accuracy,
                                    bin_density=10000, export_pathname=None):
    print("\nComputing extinction coefficients ... ", end='')

    # splitting use the ray distances into bins (max ray distance in RayCasting is 2x the max cube diagonal)
    bins = np.linspace(0, np.sqrt(ws.len_x()**2 + ws.len_y()**2 + ws.len_z()**2), bin_density, dtype=float)
    bins_midpoints = (bins[:-1] + bins[1:]) / 2.

    if export_pathname is not None:
        dirs = ['x', 'y', 'z']
        fig, ax = plt.subplots(1, 3, figsize=(13, 4))
        fig.tight_layout(pad=4, w_pad=3, h_pad=0)

    beta_out = [None, None, None]
    beta_std_out = [None, None, None]

    # binning the ray distances
    for dim in range(3):

        # probability density function of a ray travelling a certain distance
        pdf, _ = np.histogram(rays_distances[:, dim], bins=bins)
        pdf = pdf.astype(float) / np.sum(pdf)

        # integrating the pdf into a cdf
        cdf = np.cumsum(pdf)
        exp_curve_rays = 1. - cdf

        beta, cov = curve_fit(f=lambda x, b: np.exp(-b * x), xdata=bins_midpoints, ydata=exp_curve_rays)
        beta_std = np.sqrt(np.diag(cov))
        exp_curve_fit = np.exp(-beta * bins)

        beta_out[dim] = beta[0] / ws.voxel_length
        beta_std_out[dim] = beta_std[0] / ws.voxel_length

        if export_pathname is not None:
            ax[dim].plot(bins_midpoints, exp_curve_rays, 'k-', label="Ray lengths " + dirs[dim])
            ax[dim].plot(bins, exp_curve_fit, 'r-', label="Exp. fit " + dirs[dim])
            ax[dim].set_xlim(0, 0.7 * np.sqrt(ws.len_x()**2 + ws.len_y()**2 + ws.len_z()**2))
            ax[dim].set_ylim(0, 1)
            ax[dim].set_aspect(1./ax[dim].get_data_ratio())
            ax[dim].set_xlabel("Ray length (voxels)")
            ax[dim].set_ylabel("I(s)/I_0")
            ax[dim].grid()
            ax[dim].legend(loc="upper right")
            ax[dim].set_title("beta: {:.1f} +/- {:.1f}".format(beta_out[dim], beta_std_out[dim]))

    if export_pathname is not None:
        fig.suptitle("Radiation simulation with {} sources and {}° between rays"
                     .format(sources_number, degree_accuracy))
        fig.savefig(export_pathname)
    print("Done")
    return beta_out, beta_std_out



import sys
import os
import numpy as np
from pyevtk.hl import pointsToVTK
import itertools
from pumapy import Workspace


class RayCasting:

    def __init__(self, workspace, degree_accuracy, source_locations, valid_phase, boundary_behavior=0,
                 exportparticles_filepathname=''):
        self.ws = workspace
        self.degree_accuracy = degree_accuracy
        self.source_locations = source_locations
        self.valid_phase = valid_phase
        self.boundary_behavior = boundary_behavior  # 0=kill particles, 1=periodic BC
        self.exportparticles_filepathname = exportparticles_filepathname

        self.particles_number = int((180. / self.degree_accuracy - 1) * (360. / self.degree_accuracy) + 2)
        print("Number of particles in Ray Tracing simulation: {}".format(self.particles_number))

        self.rays_distances = None
        self.spherical_walkers = None
        self.X = None
        self.Y = None
        self.Z = None
        self.counter = itertools.count()

    def generate_spherical_walkers(self):
        # spherical_walkers arranged as:
        # dir=0,1,2; vox=3,4,5; pos=6,7,8; face_intersected=9; intersection=10, distance=11, particleID=12
        # x_distance=13, y_distance=14, z_distance=15
        self.spherical_walkers = np.zeros((self.particles_number, 16))

        # top and bottom particles in sphere
        self.spherical_walkers[[0, -1], 0] = np.sin([0, np.pi]) * np.cos([0, 2. * np.pi])  # dir_x
        self.spherical_walkers[[0, -1], 1] = np.sin([0, np.pi]) * np.sin([0, 2. * np.pi])  # dir_y
        self.spherical_walkers[[0, -1], 2] = np.cos([0, np.pi])  # dir_z

        # uniformly sampling a sphere by angles theta (arc from 0-180) and phi (circumference)
        arc_particles = int(180 / self.degree_accuracy) - 1
        arcs = int(360 / self.degree_accuracy)
        theta = np.linspace(self.degree_accuracy, 180 - self.degree_accuracy, arc_particles)
        theta = np.tile(theta, arcs)
        phi = np.linspace(0, 360 - self.degree_accuracy, arcs)
        phi = np.repeat(phi, arc_particles)
        self.spherical_walkers[1:-1, 0] = np.sin(theta * (np.pi/180.)) * np.cos(phi * (np.pi/180.))  # dir_x
        self.spherical_walkers[1:-1, 1] = np.sin(theta * (np.pi/180.)) * np.sin(phi * (np.pi/180.))  # dir_y
        self.spherical_walkers[1:-1, 2] = np.cos(theta * (np.pi/180.))  # dir_z

        # give particles IDs
        self.spherical_walkers[:, 12] = np.arange(self.particles_number)

    def expand_sources(self):
        # iterative loop through sources
        for i in range(self.source_locations.shape[0]):
            sys.stdout.write("\rShooting particles from sources ... {:.1f}% "
                             .format(max(i, 1) / max((self.source_locations.shape[0] - 1), 1) * 100))
            self.__reset_walkers(i)
            self.__execute_walks(i)
            self.rays_distances[i * self.particles_number:(i + 1) * self.particles_number] = self.spherical_walkers[:, 13:]
        print("Done")

    def __reset_walkers(self, i):
        # reset walkers positions to the center of source voxel
        self.spherical_walkers[:, 3:6] = self.source_locations[i]  # vox
        self.spherical_walkers[:, 6:9] = self.source_locations[i] + 0.5  # pos
        self.spherical_walkers[:, 9:12] = np.zeros((self.particles_number, 3))  # intersection, distance
        self.spherical_walkers[:, 13:] = np.zeros((self.particles_number, 3))

    def __execute_walks(self, source_number):

        # valid particles
        valid_mask = np.ones(self.spherical_walkers.shape[0], dtype=bool)

        # max travelled distance: diagonal of rectangular prism
        max_distance_void = 2. * np.sqrt(self.X**2 + self.Y**2 + self.Z**2)

        # keep walking until an intersection
        # dir=0,1,2; vox_index=3,4,5; pos=6,7,8; face_intersected=9; intersection=10, distance=11, particleID=12,
        # x_distance=13, y_distance=14, z_distance=15
        while np.sum(valid_mask) > 0:

            # save old position to compute distance increment
            old_pos = self.spherical_walkers[:, 6:9].copy()

            # compute new position:6,7,8 and face intersected:9 and new voxel indices:3,4,5
            self.__next_face_intersected(valid_mask)

            # compute distances
            self.spherical_walkers[valid_mask, 11] += np.linalg.norm(self.spherical_walkers[valid_mask, 6:9] - old_pos[valid_mask], axis=1)
            self.spherical_walkers[valid_mask, 13] += np.abs(self.spherical_walkers[valid_mask, 6] - old_pos[valid_mask, 0])
            self.spherical_walkers[valid_mask, 14] += np.abs(self.spherical_walkers[valid_mask, 7] - old_pos[valid_mask, 1])
            self.spherical_walkers[valid_mask, 15] += np.abs(self.spherical_walkers[valid_mask, 8] - old_pos[valid_mask, 2])

            # identify particles outside domain
            mask2 = (((np.any(self.spherical_walkers[:, 3:6] < 0, axis=1)) | (self.spherical_walkers[:, 3] > self.X - 1)) |
                     (self.spherical_walkers[:, 4] > self.Y - 1)) | (self.spherical_walkers[:, 5] > self.Z - 1)

            # boundary behavior: 0:kill particle, 1:periodic domain
            if self.boundary_behavior == 1:  # periodic BC for particles: re-enter the other side
                self.spherical_walkers[mask2, 3] %= self.X  # voxel indices
                self.spherical_walkers[mask2, 4] %= self.Y
                self.spherical_walkers[mask2, 5] %= self.Z
                zeros = self.spherical_walkers[:, 6] == 0  # particle positions
                self.spherical_walkers[zeros, 6] = self.X
                self.spherical_walkers[~zeros, 6] %= self.X
                zeros = self.spherical_walkers[:, 7] == 0
                self.spherical_walkers[zeros, 7] = self.Y
                self.spherical_walkers[~zeros, 7] %= self.Y
                zeros = self.spherical_walkers[:, 8] == 0
                self.spherical_walkers[zeros, 8] = self.Z
                self.spherical_walkers[~zeros, 8] %= self.Z
            else:  # kill particles
                self.spherical_walkers[mask2, 10] = 2

            # check material of next voxels
            mask2 = (~mask2) & (valid_mask)
            self.spherical_walkers[mask2, 10] = self.ws[self.spherical_walkers[mask2, 3].astype(int),
                                                        self.spherical_walkers[mask2, 4].astype(int),
                                                        self.spherical_walkers[mask2, 5].astype(int)] != self.valid_phase

            # valid particles: not collided i.e. 10=0 and travelled distance (11) less than 2x diagonal
            valid_mask = (self.spherical_walkers[:, 10] == 0) & (self.spherical_walkers[:, 11] < max_distance_void)

            if self.exportparticles_filepathname is not '':
                mask_wallcollisions = self.spherical_walkers[:, 10] != 2
                if self.spherical_walkers[mask_wallcollisions].shape[0] != 0:
                    pointsToVTK(self.exportparticles_filepathname + "" + str(source_number) + "" + str(next(self.counter)),
                                self.spherical_walkers[mask_wallcollisions, 6].copy(),
                                self.spherical_walkers[mask_wallcollisions, 7].copy(),
                                self.spherical_walkers[mask_wallcollisions, 8].copy(),
                                data={"collision": self.spherical_walkers[mask_wallcollisions, 10].copy(),
                                      "distance": self.spherical_walkers[mask_wallcollisions, 11].copy(),
                                      "id": self.spherical_walkers[mask_wallcollisions, 12].copy()})

    def __next_face_intersected(self, valid_mask):
        # dir=0,1,2; vox=3,4,5; pos=6,7,8; face_intersected=9; intersection=10, distance=11
        n = self.spherical_walkers[:, 3:6].copy()  # next intersecting face in each direction
        t = np.full_like(n, 1e10)  # time to intersecting face in each direction

        for i in range(3):
            # Step 1: Setting nX, nY, and nZ to be the next interfaces reached for each direction
            n[(self.spherical_walkers[:, i] > 0) & (valid_mask), i] += 1

            # Step 2: Setting tX, tY, and tZ to be the time to reach the next interface in each direction
            mask = (self.spherical_walkers[:, i] != 0) & (valid_mask)
            t[mask, i] = (n[mask, i] - self.spherical_walkers[mask, 6 + i]) / self.spherical_walkers[mask, i]

        # Step 3: Setting the next position of the spherical_walker based on the nearest interface
        mask = ((t[:, 0] <= t[:, 1]) & (t[:, 0] <= t[:, 2])) & (valid_mask)
        self.spherical_walkers[mask, 6:10] = np.column_stack((n[mask, 0],
                                                              self.spherical_walkers[mask, 7] + self.spherical_walkers[mask, 1] * t[mask, 0],
                                                              self.spherical_walkers[mask, 8] + self.spherical_walkers[mask, 2] * t[mask, 0],
                                                              np.zeros((np.sum(mask), 1))))

        mask2 = ((t[:, 1] <= t[:, 2]) & (t[:, 1] <= t[:, 0])) & (valid_mask)
        self.spherical_walkers[mask2, 6:10] = np.column_stack((self.spherical_walkers[mask2, 6] + self.spherical_walkers[mask2, 0] * t[mask2, 1],
                                                               n[mask2, 1],
                                                               self.spherical_walkers[mask2, 8] + self.spherical_walkers[mask2, 2] * t[mask2, 1],
                                                               np.ones((np.sum(mask2), 1))))

        mask = ((~mask) & (~mask2)) & (valid_mask)
        self.spherical_walkers[mask, 6:10] = np.column_stack((self.spherical_walkers[mask, 6] + self.spherical_walkers[mask, 0] * t[mask, 2],
                                                              self.spherical_walkers[mask, 7] + self.spherical_walkers[mask, 1] * t[mask, 2],
                                                              n[mask, 2],
                                                              np.full((np.sum(mask), 1), 2)))

        # update voxel indices
        for i in range(3):
            mask = ((self.spherical_walkers[:, 9] == np.arange(3)[i]) & (self.spherical_walkers[:, i] > 0)) & (valid_mask)
            self.spherical_walkers[mask, 3 + i] += 1
            mask = ((self.spherical_walkers[:, 9] == np.arange(3)[i]) & (self.spherical_walkers[:, i] <= 0)) & (valid_mask)
            self.spherical_walkers[mask, 3 + i] -= 1

    def error_check(self):
        if not isinstance(self.ws, Workspace):
            raise Exception("Workspace must be a puma.Workspace.")
        else:
            self.X = self.ws.matrix.shape[0]
            self.Y = self.ws.matrix.shape[1]
            self.Z = self.ws.matrix.shape[2]

        if 180 % self.degree_accuracy != 0:
            raise Exception("Ray separation can only be an exact divider of 180°")

        if not isinstance(self.source_locations, np.ndarray) or self.source_locations.shape[1] != 3:
            raise Exception("Source locations has to be a Numpy array of shape (NumberOfSource, 3).")
        self.rays_distances = np.zeros((self.source_locations.shape[0] * self.particles_number, 3))

        if self.exportparticles_filepathname is not '':
            if not os.path.exists(os.path.split(self.exportparticles_filepathname)[0]):
                raise Exception("Directory " + os.path.split(self.exportparticles_filepathname)[0] + " not found.")

        if np.count_nonzero(self.ws.matrix == self.valid_phase) == 0:
            raise Exception("No valid voxels detected (i.e. ID={}), cannot run radiation ray tracing.".format(self.valid_phase))



from pumapy.utilities.generic_checks import check_ws_cutoff
from pumapy.utilities.logger import print_warning
import numpy as np


def compute_mean_intercept_length(workspace, void_cutoff):
    """ Computation of the mean intercept length

    :param workspace: domain
    :type workspace: Workspace
    :param void_cutoff: specify the void or gaseous phase of the domain
    :type void_cutoff: tuple(int, int)
    :return: mean intercept length in x,y,z
    :rtype: tuple(float, float, float)
    """

    check_ws_cutoff(workspace, void_cutoff)
    workspace.log.log_section("Computing mean intercept length")
    workspace.log.log_line("Domain Size: " + str(workspace.get_shape()))
    workspace.log.log_line("Void cutoff: " + str(void_cutoff))

    ws = workspace.copy()
    ws.binarize_range(void_cutoff)
    void_voxs = ws.matrix.sum() * ws.voxel_length

    ws.matrix = ws.matrix.astype(int)

    if void_voxs > 0:
        collisions_x = np.count_nonzero(ws.matrix[:-1] - ws.matrix[1:] > 0)
        collisions_y = np.count_nonzero(ws.matrix[:, :-1] - ws.matrix[:, 1:] > 0)
        collisions_z = np.count_nonzero(ws.matrix[:, :, :-1] - ws.matrix[:, :, 1:] > 0)

        mil = []

        if collisions_x == 0:
            print_warning("Infinite Mean Intercept Length in x direction")
            mil.append(float('Inf'))
        else:
            mil.append(void_voxs / collisions_x)
        if collisions_y == 0:
            print_warning("Infinite Mean Intercept Length in y direction")
            mil.append(float('Inf'))
        else:
            mil.append(void_voxs / collisions_y)
        if collisions_z == 0:
            print_warning("Infinite Mean Intercept Length in z direction")
            mil.append(float('Inf'))
        else:
            mil.append(void_voxs / collisions_z)
    else:
        print_warning("Domain is fully dense.")
        mil = 0, 0, 0

    workspace.log.log_line("Mean Intercept Length: " + str(mil))
    workspace.log.write_log()
    return tuple(mil)



from pyevtk.hl import imageToVTK
import numpy as np
from skimage.io import imsave
import pickle
from pumapy.utilities.workspace import Workspace
from pumapy.utilities.logger import print_warning
from pumapy.utilities.isosurface import generate_isosurface
from os import path


def export_vti(filename, dict_data, voxel_length=None):
    """ Export either a puma.Workspace or numpy array to vti

    :param filename: filepath and name
    :type filename: string
    :param dict_data: dictionary setup as {"name1": data1, "name2": data2 ...} containing either Workspaces or ndarrays
    :type dict_data: dict
    :param voxel_length: with voxel length to give to Numpy arrays (if any)
    :type voxel_length: float, optional
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws_vtk = puma.import_vti(puma.path_to_example_file("fibers_with_orientation.vti"))
    >>> puma.export_vti("fibers_with_orientation.vti", ws_vtk)
    """

    # error path checks
    if not isinstance(filename, str):
        raise Exception("Filename has to be a string.")
    filename_split = path.split(filename)
    if filename_split[0] != '' and not path.exists(path.split(filename)[0]):
        raise Exception("Directory " + filename_split[0] + " not found.")
    if filename[-4:] == ".vti":
        filename = filename[:-4]

    print("Exporting " + filename + ".vti ... ", end='')

    if isinstance(dict_data, Workspace) or isinstance(dict_data, np.ndarray):
        dict_data = {"data": dict_data}

    dict_to_export = dict()

    for name, data in dict_data.items():

        if isinstance(data, Workspace):
            mat = data.matrix.copy()  # need to copy to make sure data is "continuous"
            dict_to_export[name] = mat

            if data.orientation.shape[:3] == data.matrix.shape:
                orientation = (data.orientation[:, :, :, 0].copy(),
                               data.orientation[:, :, :, 1].copy(),
                               data.orientation[:, :, :, 2].copy())
                dict_to_export[name] = mat
                dict_to_export[name + "_orient"] = orientation

            if voxel_length is None:
                voxel_length = data.voxel_length

        elif isinstance(data, np.ndarray):
            if voxel_length is None:
                voxel_length = 1e-6
            if data.ndim == 2:  # 2D numpy array to 3D
                data = np.expand_dims(data.copy(), axis=2)
            elif data.ndim == 3:
                data = data.copy()
            elif data.ndim == 4 and data.shape[3] == 3:
                data = (data[:, :, :, 0].copy(),
                        data[:, :, :, 1].copy(),
                        data[:, :, :, 2].copy())
            else:
                raise Exception("Numpy array has to be either 3D for scalars or 4D with shape[3]=3.")

            dict_to_export[name] = data
        else:
            raise Exception("Data to export to vti needs to be either a pumapy.Workspace or Numpy array")

    imageToVTK(filename, spacing=[voxel_length] * 3, cellData=dict_to_export)

    print("Done")
    return True


def export_3Dtiff(filename, ws_or_nparray, to8bit=False):
    """ Export either a puma.Workspace or numpy array to 3Dtiff

    :param filename: filepath and name
    :type filename: string
    :param ws_or_nparray: to be exported
    :type ws_or_nparray: Workspace or ndarray
    :param to8bit: if True, it converts the image to 8bit, otherwise 16bit is exported
    :type to8bit: bool, optional
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws_tiff = puma.import_3Dtiff(puma.path_to_example_file("50_artfibers.tif"), 1.3e-6, import_ws=True)
    >>> puma.export_3Dtiff("50_artfibers.tif", ws_tiff)
    """

    # error checks
    if not isinstance(filename, str):
        raise Exception("Filename has to be a string.")
    filename_split = path.split(filename)
    if filename_split[0] != '' and not path.exists(path.split(filename)[0]):
        raise Exception("Directory " + filename_split[0] + " not found.")
    if isinstance(ws_or_nparray, Workspace):
        data = ws_or_nparray.matrix.copy()
    elif isinstance(ws_or_nparray, np.ndarray):
        data = ws_or_nparray.copy()
    else:
        raise Exception("Data to export to 3Dtiff needs to be either a pumapy.Workspace or Numpy array")

    if filename[-4:] != '.tif' and filename[-5:] != '.tiff':
        filename += '.tif'

    if to8bit:
        if data.max() > 255:
            print_warning("Data max is more than 255, normalizing to range 0-255.")
            data *= 255.0 / data.max()
        data = data.astype(np.uint8)
    else:
        data = data.astype(np.uint16)

    print("Exporting " + filename + " ... ", end='')
    if data.ndim == 2:
        imsave(filename, data.transpose((1, 0)), check_contrast=False)
    else:
        imsave(filename, data.transpose((2, 1, 0)), check_contrast=False)
    print("Done")
    return True


def export_bin(filename, ws):
    """ Export a puma.Workspace to binary (.pumapy extension)

    :param filename: filepath and name
    :type filename: string
    :param ws: to be exported
    :type: Workspace
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws_binary = puma.import_bin(puma.path_to_example_file("fibers_with_orientation.pumapy"))
    >>> puma.export_bin("fibers_with_orientation.vti", ws_binary)
    """

    # error checks
    if not isinstance(filename, str):
        raise Exception("Filename has to be a string.")
    filename_split = path.split(filename)
    if filename_split[0] != '' and not path.exists(path.split(filename)[0]):
        raise Exception("Directory " + filename_split[0] + " not found.")
    if not isinstance(ws, Workspace):
        raise Exception("Data to export to bin needs to be a pumapy.Workspace")

    if filename[-7:] != ".pumapy":
        filename += '.pumapy'

    print("Exporting " + filename + " ... ", end='')
    output = open(filename, 'wb')
    pickle.dump(ws, output)
    output.close()
    print("Done")
    return True


def export_sparta_implicit_surfaces(filename, ws):
    """ Export a puma.Workspace to binary (.pumapy extension)

    :param filename: filepath and name
    :type filename: string
    :param ws: to be exported
    :type ws: Workspace
    :return: True if successful, False otherwise.
    :rtype: bool
    """

    # error checks
    if not isinstance(filename, str):
        raise Exception("Filename has to be a string.")
    filename_split = path.split(filename)
    if filename_split[0] != '' and not path.exists(path.split(filename)[0]):
        raise Exception("Directory " + filename_split[0] + " not found.")
    if not isinstance(ws, Workspace):
        raise Exception("Data to export to bin needs to be a pumapy.Workspace")

    if filename[-7:] != ".pumapy.isurf":
        filename += '.pumapy.isurf'

    print("Exporting " + filename + " ... ", end='')

    if ws.len_z() == 1:
        f = open(filename, 'w+b')
        byte_arr = [ws.len_x(), ws.len_y()]
        binary_format = bytearray(byte_arr)
        f.write(binary_format)
        pickle.dump(ws.matrix, f)
        f.close()

    print("Done")
    return True


def export_stl(filename, ws, cutoff, flag_closed_edges=True, flag_gaussian=False, binary=True):
    """ Export a puma.Workspace to STL

    :param filename: filepath and name
    :type filename: string
    :param ws: to be exported
    :type ws: Workspace
    :param cutoff: specify cutoff to binarize material
    :type cutoff: tuple(int, int)
    :param flag_closed_edges: close the surface edges on the boundaries
    :type flag_closed_edges: bool, optional
    :param flag_gaussian: apply Gaussian filter before creating surface
    :type flag_gaussian: bool, optional
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws_imported = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
    >>> puma.export_stl('200_fiberform', ws_imported, cutoff=(100, 255), flag_closed_edges=True)
    """

    # error checks
    if not isinstance(filename, str):
        raise Exception("Filename has to be a string.")
    filename_split = path.split(filename)
    if filename_split[0] != '' and not path.exists(path.split(filename)[0]):
        raise Exception("Directory " + filename_split[0] + " not found.")
    if not isinstance(ws, Workspace):
        raise Exception("Data to export to stl needs to be a pumapy.Workspace")

    if filename[-4:] != ".stl":
        filename += '.stl'

    print("Exporting " + filename + " ... ", end='')
    mesh = generate_isosurface(ws, cutoff, flag_closed_edges, flag_gaussian)
    mesh.save(filename, binary)
    print("Done")
    return True



import time


class Timer:
    current = 0

    def __init__(self):
        self.current = time.perf_counter()

    def start(self):
        self.current = time.perf_counter()

    def reset(self):
        self.current = time.perf_counter()

    def elapsed(self):
        return time.perf_counter() - self.current

    def print_elapsed(self, msg=""):
        print(msg, (time.perf_counter() - self.current))



"""
Copyright @ 2017, 2020, 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.
This software may be used, reproduced, and provided to others only as permitted under the terms of the agreement under which it was acquired from the U.S. Government. Neither title to, nor ownership of, the software is hereby transferred. This notice shall remain on all copies of the software.
This file is available under the terms of the NASA Open Source Agreement (NOSA), and further subject to the additional disclaimer below:
Disclaimer:
THE SOFTWARE AND/OR TECHNICAL DATA ARE PROVIDED "AS IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY THAT THE SOFTWARE AND/OR TECHNICAL DATA WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SOFTWARE AND/OR TECHNICAL DATA WILL BE ERROR FREE, OR ANY WARRANTY THAT TECHNICAL DATA, IF PROVIDED, WILL CONFORM TO THE SOFTWARE. IN NO EVENT SHALL THE UNITED STATES GOVERNMENT, OR ITS CONTRACTORS OR SUBCONTRACTORS, BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT, INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM, OR IN ANY WAY CONNECTED WITH THIS SOFTWARE AND/OR TECHNICAL DATA, WHETHER OR NOT BASED UPON WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED BY PERSONS OR PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE AND/OR TECHNICAL DATA.
THE UNITED STATES GOVERNMENT DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD PARTY COMPUTER SOFTWARE, DATA, OR DOCUMENTATION, IF SAID THIRD PARTY COMPUTER SOFTWARE, DATA, OR DOCUMENTATION IS PRESENT IN THE NASA SOFTWARE AND/OR TECHNICAL DATA, AND DISTRIBUTES IT "AS IS."
RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE UNITED STATES GOVERNMENT AND ITS CONTRACTORS AND SUBCONTRACTORS, AND SHALL INDEMNIFY AND HOLD HARMLESS THE UNITED STATES GOVERNMENT AND ITS CONTRACTORS AND SUBCONTRACTORS FOR ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES THAT MAY ARISE FROM RECIPIENT'S USE OF THE SOFTWARE AND/OR TECHNICAL DATA, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, THE USE THEREOF.
IF RECIPIENT FURTHER RELEASES OR DISTRIBUTES THE NASA SOFTWARE AND/OR TECHNICAL DATA, RECIPIENT AGREES TO OBTAIN THIS IDENTICAL WAIVER OF CLAIMS, INDEMNIFICATION AND HOLD HARMLESS, AGREEMENT WITH ANY ENTITIES THAT ARE PROVIDED WITH THE SOFTWARE AND/OR TECHNICAL DATA.
"""

""" pumapy
Root directory for the pumapy package.
"""

# utilities
from pumapy.utilities.workspace import Workspace
from pumapy.utilities.logger import Logger, print_warning
from pumapy.utilities.timer import Timer
from pumapy.utilities.isosurface import generate_isosurface
from pumapy.utilities.property_maps import IsotropicConductivityMap, AnisotropicConductivityMap, ElasticityMap
from pumapy.utilities.boundary_conditions import ConductivityBC, ElasticityBC
from pumapy.utilities.example_files import path_to_example_file, list_example_files

# input/output
from pumapy.io.input import import_3Dtiff, import_bin, import_weave_vtu, import_vti
from pumapy.io.output import export_vti, export_3Dtiff, export_bin, export_sparta_implicit_surfaces, export_stl
try:
    from pumapy.io.export_texgen_weave import export_weave_vtu
except:
    print_warning("WARNING: 'import TexGen.Core' failed: cannot use TexGen functions and pumapy.export_weave_vtu.")

# material properties
from pumapy.materialproperties.surfacearea import compute_surface_area
from pumapy.materialproperties.volumefraction import compute_volume_fraction
from pumapy.materialproperties.mean_intercept_length import compute_mean_intercept_length
from pumapy.materialproperties.orientation import compute_orientation_st, compute_angular_differences
from pumapy.materialproperties.conductivity import compute_thermal_conductivity, compute_electrical_conductivity
from pumapy.materialproperties.tortuosity import compute_continuum_tortuosity
from pumapy.materialproperties.elasticity import compute_elasticity, compute_stress_analysis
from pumapy.materialproperties.radiation import compute_radiation, compute_extinction_coefficients
from pumapy.materialproperties.permeability import compute_permeability

# filtering
from pumapy.filters.filters import (filter_median, filter_gaussian, filter_edt, filter_mean,
                                    filter_erode, filter_dilate, filter_opening, filter_closing)

# generation
from pumapy.generation.tpms import generate_tpms
from pumapy.generation.sphere import get_sphere
from pumapy.generation.random_spheres import generate_random_spheres
from pumapy.generation.generate_sphere import generate_sphere
from pumapy.generation.generate_2d_square_array import generate_2d_square_array
from pumapy.generation.random_fibers import generate_random_fibers
try:
    from pumapy.generation.weave_3mdcp.weave_3mdcp import generate_3mdcp
except ImportError:  # import it only if installed
    pass

# visualization
from pumapy.visualization.render import render_volume, render_contour, render_orientation, render_contour_multiphase
from pumapy.visualization.slicer import plot_slices, compare_slices

# segmentation
from pumapy.segmentation.porespace import identify_porespace, fill_closed_pores



import unittest
import numpy as np
import pumapy as puma


class TestImportExportVTKI(unittest.TestCase):

    def test_io_vti_ws(self):
        ws = puma.Workspace.from_shape_value((10, 11, 12), 34)
        ws.matrix[0, -1, 0] = 1
        ws.matrix[0, 0, -1] = 2
        ws.matrix[-1, -1, -1] = 3
        ws.matrix[-1, 0, 0] = 4
        np.testing.assert_equal(puma.export_vti("out/ws1", ws), True)
        ws2 = ws.copy()
        ws = puma.import_vti("out/ws1.vti")
        np.testing.assert_equal(ws.matrix, ws2.matrix)

    def test_io_vti_nparray(self):
        nparray = np.full((10, 11, 12), 34, dtype=np.uint16)
        nparray[0, -1, 0] = 1
        nparray[0, 0, -1] = 2
        nparray[-1, -1, -1] = 3
        nparray[-1, 0, 0] = 4
        np.testing.assert_equal(puma.export_vti("out/nparray1", nparray), True)
        nparray2 = nparray.copy()
        nparray = puma.import_vti("out/nparray1.vti", import_ws=False)["data"]
        np.testing.assert_equal(nparray, nparray2)

    def test_io_vti_2Dnparray(self):
        nparray = np.full((10, 11), 34, dtype=np.uint16)
        nparray[0, -1] = 1
        nparray[-1, -1] = 3
        nparray[-1, 0] = 4
        np.testing.assert_equal(puma.export_vti("out/nparray2", nparray), True)
        nparray2 = np.expand_dims(nparray.copy(), axis=2)
        nparray = puma.import_vti("out/nparray2.vti", import_ws=False)["data"]
        np.testing.assert_equal(nparray, nparray2)

    def test_io_vti_nparray_orientation(self):
        nparray = np.tile((1, 0, 0), (5, 10, 20, 1)).astype(float)
        nparray[0, -1, 0] = [0, 1, 0]
        nparray[0, 0, -1] = [0, 0, 1]
        nparray[-1, -1, -1] = [10, 10, 10]
        nparray[-1, 0, 0] = [10, 0, 1]
        np.testing.assert_equal(puma.export_vti("out/nparray3", nparray), True)
        nparray2 = nparray.copy()
        nparray = puma.import_vti("out/nparray3.vti", import_ws=False)["data"]
        np.testing.assert_equal(nparray, nparray2)

    def test_io_vti_orientation(self):
        ws = puma.Workspace.from_shape_value_vector((5, 10, 20), 47, (1, 0, 0))
        ws.orientation[0, -1, 0] = [0, 1, 0]
        ws.orientation[0, 0, -1] = [0, 0, 1]
        ws.orientation[-1, -1, -1] = [10, 10, 10]
        ws.orientation[-1, 0, 0] = [10, 0, 1]
        np.testing.assert_equal(puma.export_vti("out/ws2", ws), True)
        ws2 = ws.copy()
        ws = puma.import_vti("out/ws2.vti")
        np.testing.assert_equal(ws.orientation, ws2.orientation)
        np.testing.assert_equal(ws.matrix, ws2.matrix)

    def test_io_vtk_ws(self):
        # Created as:
        # puma::MatVec3<double> dirs(10, 11, 7, puma::Vec3<double>(1, 0, 0));
        # dirs.at(1, 2, 3) = puma::Vec3<double>(1, 2, 3);
        # dirs.at(7, 6, 5) = puma::Vec3<double>(7, 6, 5);
        # dirs.at(3, 10, 2) = puma::Vec3<double>(3, 10, 2);
        orient_test = puma.import_vti("testdata/orient_test.vtk", import_ws=False)["PuMA_Vec3"]
        nparray = np.tile((1, 0, 0), (10, 11, 7, 1)).astype(float)
        nparray[1, 2, 3] = [1, 2, 3]
        nparray[7, 6, 5] = [7, 6, 5]
        nparray[3, 10, 2] = [3, 10, 2]
        np.testing.assert_equal(orient_test, nparray)


class TestImportExport3Dtiff(unittest.TestCase):

    def test_export_3Dtiff_ws(self):
        ws = puma.Workspace.from_shape_value((10, 11, 12), 34)
        ws.matrix[0, -1, 0] = 1
        ws.matrix[0, 0, -1] = 2
        ws.matrix[-1, -1, -1] = 3
        ws.matrix[-1, 0, 0] = 4
        np.testing.assert_equal(puma.export_3Dtiff("out/ws", ws), True)

    def test_import_3Dtiff_ws(self):
        ws = puma.import_3Dtiff("out/ws.tif")
        ws2 = puma.Workspace.from_shape_value((10, 11, 12), 34)
        ws2.matrix[0, -1, 0] = 1
        ws2.matrix[0, 0, -1] = 2
        ws2.matrix[-1, -1, -1] = 3
        ws2.matrix[-1, 0, 0] = 4
        np.testing.assert_equal(ws.matrix, ws2.matrix)

    def test_export_3Dtiff_nparray(self):
        nparray = np.full((10, 11, 12), 34, dtype=np.uint16)
        nparray[0, -1, 0] = 1
        nparray[0, 0, -1] = 2
        nparray[-1, -1, -1] = 3
        nparray[-1, 0, 0] = 4
        np.testing.assert_equal(puma.export_3Dtiff("out/nparray", nparray), True)

    def test_import_3Dtiff_nparray(self):
        nparray = puma.import_3Dtiff("out/nparray.tif", import_ws=False)
        nparray2 = np.full((10, 11, 12), 34, dtype=np.uint16)
        nparray2[0, -1, 0] = 1
        nparray2[0, 0, -1] = 2
        nparray2[-1, -1, -1] = 3
        nparray2[-1, 0, 0] = 4
        np.testing.assert_equal(nparray, nparray2)


class TestImportExportbin(unittest.TestCase):

    def test_export_bin_ws(self):
        ws = puma.Workspace.from_shape_value((10, 11, 12), 34)
        ws.matrix[0, -1, 0] = 1
        ws.matrix[0, 0, -1] = 2
        ws.matrix[-1, -1, -1] = 3
        ws.matrix[-1, 0, 0] = 4
        np.testing.assert_equal(puma.export_bin("out/ws", ws), True)

    def test_import_bin_ws(self):
        ws = puma.import_bin("out/ws.pumapy")
        ws2 = puma.Workspace.from_shape_value((10, 11, 12), 34)
        ws2.matrix[0, -1, 0] = 1
        ws2.matrix[0, 0, -1] = 2
        ws2.matrix[-1, -1, -1] = 3
        ws2.matrix[-1, 0, 0] = 4
        np.testing.assert_equal(ws.matrix, ws2.matrix)


if __name__ == '__main__':
    unittest.main()



import sys
import numpy as np
from scipy.ndimage.filters import gaussian_filter
from scipy.ndimage import distance_transform_edt
from pumapy.utilities.generic_checks import check_ws_cutoff


def compute_angular_differences(matrix, orientation1, orientation2, cutoff):
    """ Compute angular difference between two orientation ndarrays

    :param matrix: domain matrix
    :type matrix: ndarray
    :param orientation1: orientation as (x, y, z, 3)
    :type orientation1: ndarray
    :param orientation2: orientation as (x, y, z, 3)
    :type orientation2: ndarray
    :param cutoff: to binarize domain
    :type cutoff: tuple(int, int) 
    :return: angle_errors in degrees, mean, std
    :rtype: tuple(ndarray, float, float)
    """

    if not isinstance(matrix, np.ndarray) or not isinstance(orientation1, np.ndarray) or not isinstance(orientation2, np.ndarray):
        raise Exception("Inputs must be ndarrays.")
    if not isinstance(cutoff, tuple) or not len(cutoff) == 2:
        raise Exception("Cutoff must be a tuple(int, int).")
    if not (orientation1.ndim == 4 and orientation2.ndim == 4 and matrix.ndim == 3 and orientation1.shape[3] == 3 and
            orientation1.shape == orientation2.shape and orientation1.shape[0] == matrix.shape[0] and
            orientation1.shape[1] == matrix.shape[1] and orientation1.shape[2] == matrix.shape[2]):
        raise Exception("Incorrect dimensions in input ndarrays.")

    mask = np.logical_and(matrix >= cutoff[0], matrix <= cutoff[1])
    unit_vectors_1 = orientation1[mask]
    unit_vectors_2 = orientation2[mask]

    radians_diff = np.einsum('ij,ij->i', unit_vectors_1, unit_vectors_2)
    diff = np.zeros((unit_vectors_1.shape[0], 2), dtype=float)
    diff[:, 0] = np.degrees(np.arccos(np.clip(radians_diff, -1, 1)))
    diff[:, 1] = 180 - diff[:, 0]
    diff = np.min(diff, axis=1)

    angle_diff = np.zeros_like(matrix, dtype=float)
    angle_diff[mask] = diff

    return angle_diff, diff.mean(), diff.std()


def compute_orientation_st(ws, sigma, rho, cutoff, edt=False):
    """ Compute orientation of the material by the structure tensor algorithm

    :param ws: domain
    :type ws: Workspace
    :param sigma: kernel size parameter for Gaussian derivatives (should be smaller than rho)
    :type sigma: float
    :param rho: kernel size parameter for Gaussian filter (should be bigger than sigma)
    :type rho: float
    :param cutoff: which grayscales to consider
    :type cutoff: tuple(int, int)
    :param edt: indicating if we need to apply Euclidean Distance Transform before computing ST
    :type edt: bool
    :return: True if successful, False otherwise.
    :rtype: bool
    """
    solver = OrientationST(ws, sigma, rho, cutoff, edt)

    solver.error_check()
    return solver.compute()


class OrientationST:

    def __init__(self, ws, sigma, rho, cutoff, edt):
        self.ws = ws.copy()
        self.ws_in = ws

        self.sigma = sigma
        self.rho = rho
        self.cutoff = cutoff
        self.edt = edt

        self.mask = None
        self.st = None

    def compute(self):
        self.__compute_structuretensor()
        self.__eigenvalue_analysis()
        return True

    def __compute_structuretensor(self):
        self.ws.matrix = self.ws.matrix.astype(int)
        self.mask = np.logical_and(self.ws.matrix >= self.cutoff[0], self.ws.matrix <= self.cutoff[1])

        sys.stdout.write("First gradient computation ... ")
        if self.edt:
            self.st = np.empty((self.ws.matrix.shape[0], self.ws.matrix.shape[1], self.ws.matrix.shape[2], 3, 3), dtype=float)
            tmp = np.empty(self.ws.matrix.shape, dtype=float)
            distance = distance_transform_edt(self.mask)
            Vx = gaussian_filter(distance, self.sigma, order=[1, 0, 0], mode='nearest')
            Vy = gaussian_filter(distance, self.sigma, order=[0, 1, 0], mode='nearest')
            Vz = gaussian_filter(distance, self.sigma, order=[0, 0, 1], mode='nearest')
        else:
            self.st = np.empty((self.ws.matrix.shape[0], self.ws.matrix.shape[1], self.ws.matrix.shape[2], 3, 3), dtype=int)
            tmp = np.empty(self.ws.matrix.shape, dtype=int)
            Vx = gaussian_filter(self.ws.matrix, self.sigma, order=[1, 0, 0], mode='nearest')
            Vy = gaussian_filter(self.ws.matrix, self.sigma, order=[0, 1, 0], mode='nearest')
            Vz = gaussian_filter(self.ws.matrix, self.sigma, order=[0, 0, 1], mode='nearest')
        print("Done")

        # blur the gradient products with a Gaussian filter and assemble structure tensor
        sys.stdout.write("Blurring of gradients ... ")
        np.multiply(Vx, Vx, out=tmp)
        self.st[:, :, :, 0, 0] = gaussian_filter(tmp, self.rho, mode='nearest')
        np.multiply(Vy, Vy, out=tmp)
        self.st[:, :, :, 1, 1] = gaussian_filter(tmp, self.rho, mode='nearest')
        np.multiply(Vz, Vz, out=tmp)
        self.st[:, :, :, 2, 2] = gaussian_filter(tmp, self.rho, mode='nearest')
        np.multiply(Vx, Vy, out=tmp)
        self.st[:, :, :, 0, 1] = gaussian_filter(tmp, self.rho, mode='nearest')
        self.st[:, :, :, 1, 0] = self.st[:, :, :, 0, 1]
        np.multiply(Vx, Vz, out=tmp)
        self.st[:, :, :, 0, 2] = gaussian_filter(tmp, self.rho, mode='nearest')
        self.st[:, :, :, 2, 0] = self.st[:, :, :, 0, 2]
        np.multiply(Vy, Vz, out=tmp)
        self.st[:, :, :, 1, 2] = gaussian_filter(tmp, self.rho, mode='nearest')
        self.st[:, :, :, 2, 1] = self.st[:, :, :, 1, 2]
        print("Done")

    def __eigenvalue_analysis(self):
        sys.stdout.write("Computing eigenvalue analysis ... ")
        self.ws_in.set_orientation(np.zeros((self.ws.matrix.shape[0], self.ws.matrix.shape[1], self.ws.matrix.shape[2], 3), dtype=float))
        self.ws_in.orientation[self.mask] = np.linalg.eigh(self.st[self.mask])[1][:, :, 0]
        print("Done")

    def error_check(self):
        check_ws_cutoff(self.ws, self.cutoff)



from pumapy import Workspace
from pumapy.utilities.timer import Timer
import pumapy.utilities.generic_checks as check
from pumapy.generation.sphere import get_sphere
import numpy as np


def generate_sphere(size, center, diameter):
    """ Generation of a sphere at a given point and diameter

    :param size: size of 3D domain (x,y,z)
    :type size: tuple(int, int, int)
    :param center: centerpoint of sphere (x,y,z)
    :type center: tuple(int, int, int)
    :param diameter: diameter of the random spheres in voxels
    :rtype: float
    """

    generator = GeneratorSphere(size, center, diameter)

    generator.error_check()

    generator.log_input()
    generator.log_output()

    return generator.generate()


class GeneratorSphere:
    def __init__(self, size, center, diameter):
        self._workspace = Workspace()
        self._size = size
        self._center = center
        self._diameter = diameter

        self._size_padded = (self._size[0]+diameter, self._size[1]+diameter, self._size[2]+diameter)
        self._start = int(0.5 * diameter)
        self._endX = size[0] + self._start
        self._endY = size[1] + self._start
        self._endZ = size[2] + self._start

        # size of the created sphere block
        self._spheresize = int(self._diameter+4)
        if self._spheresize % 2 == 0:
            self._size_m = self._spheresize / 2
            self._size_p = self._spheresize / 2
        else:
            self._size_m = int(self._spheresize / 2.)+1
            self._size_p = int(self._spheresize / 2.)

    def generate(self):
        t = Timer()
        self._workspace.resize_new_matrix(self._size_padded)
        self._generate_sphere()
        print("Generated in: " + str(t.elapsed()) + " seconds")
        return self._workspace

    def _generate_sphere(self):
        sphere = get_sphere(self._diameter)
        [x,y,z] = self._center + np.array([self._start, self._start, self._start])
        xm = int(x - self._size_m)
        xp = int(x + self._size_p)
        ym = int(y - self._size_m)
        yp = int(y + self._size_p)
        zm = int(z - self._size_m)
        zp = int(z + self._size_p)

        self._workspace.matrix[xm:xp,ym:yp,zm:zp] += sphere[:,:,:]
        self._workspace.matrix = self._workspace.matrix[self._start:self._endX,self._start:self._endY,self._start:self._endZ]

    def log_input(self):
        self._workspace.log.log_section("Generating Sphere")
        self._workspace.log.log_value("Domain Size", self._size)
        self._workspace.log.log_value("Sphere Center", self._center)
        self._workspace.log.log_value("Diameter", self._diameter)
        self._workspace.log.write_log()

    def log_output(self):
        self._workspace.log.log_section("Finished Sphere Generation")
        self._workspace.log.write_log()

    def error_check(self):
        check.size_check(self._size)
        if self._center[0] < 0 or self._center[0] >= self._size[0]:
            raise Exception("Error, invalid center. Must be contained within domain size")
        if self._center[1] < 0 or self._center[1] >= self._size[1]:
            raise Exception("Error, invalid center. Must be contained within domain size")
        if self._center[2] < 0 or self._center[2] >= self._size[2]:
            raise Exception("Error, invalid center. Must be contained within domain size")



import numpy as np
from pumapy.utilities.generic_checks import check_ws_cutoff


def compute_volume_fraction(workspace, cutoff):
    """ Compute the volume fraction

    :param workspace: domain
    :type workspace: Workspace
    :param cutoff: to binarize domain
    :type cutoff: tuple(int, int)
    :return: volume fraction
    :rtype: float
    """
    volume_fraction = VolumeFraction(workspace, cutoff)

    volume_fraction.error_check()

    volume_fraction.log_input()
    volume_fraction.compute()
    volume_fraction.log_output()

    return volume_fraction.vf


class VolumeFraction:

    def __init__(self, workspace, cutoff):
        self.workspace = workspace
        self.cutoff = cutoff
        self.vf = -1.

    def compute(self):
        mask = self.workspace.matrix >= self.cutoff[0]
        mask_high = self.workspace.matrix <= self.cutoff[1]
        mask = mask * mask_high
        self.vf = float(np.sum(mask)) / float(self.workspace.get_size())

    def error_check(self):
        check_ws_cutoff(self.workspace, self.cutoff)

    def log_input(self):
        self.workspace.log.log_section("Computing Volume Fraction")
        self.workspace.log.log_line("Domain Size: " + str(self.workspace.get_shape))
        self.workspace.log.log_line("Cutoff: " + str(self.cutoff))
        self.workspace.log.write_log()

    def log_output(self):
        self.workspace.log.log_section("Finished Volume Fraction Calculation")
        self.workspace.log.log_line("Volume Fraction: " + str(self.vf))



from pumapy.materialproperties.volumefraction import compute_volume_fraction
from pumapy import IsotropicConductivityMap
from pumapy.physicsmodels.isotropic_conductivity import IsotropicConductivity


def compute_continuum_tortuosity(workspace, cutoff, direction, side_bc='p', prescribed_bc=None,
                                 tolerance=1e-4, maxiter=10000, solver_type='cg', display_iter=True):
    """ Compute the tortuosity modelling the local conductivity as isotropic

    :param workspace: domain
    :type workspace: Workspace
    :param cutoff: to binarize domain
    :type cutoff: tuple(int, int)
    :param direction: direction for solve ('x','y', or 'z')
    :type direction: string
    :param side_bc: side boundary conditions (string) can be symmetric ('s'), periodic ('p') or dirichlet ('d')
    :type side_bc: string
    :param prescribed_bc: 3D array holding dirichlet BC
    :type prescribed_bc: ConductivityBC
    :param tolerance: tolerance for iterative solver
    :type tolerance: float
    :param maxiter: maximum Iterations for solver
    :type maxiter: int
    :param solver_type: solver type, options: 'cg', 'bicgstab', 'direct'
    :type solver_type: string
    :param display_iter: display iterations and residual
    :type display_iter: bool
    :return: tortuosity, diffusivity, porosity, concentration field
    :rtype: tuple(tuple(float, float, float), float, float, ndarrya)
    """

    cond_map = IsotropicConductivityMap()
    cond_map.add_material(cutoff, 1)

    if cutoff[0] > 0:
        cond_map.add_material((0, cutoff[0]-1),0)
    cond_map.add_material((cutoff[1]+1,32000),0)


    solver = IsotropicConductivity(workspace, cond_map, direction, side_bc, prescribed_bc,
                                   tolerance, maxiter, solver_type, display_iter)

    solver.error_check()

    solver.log_input()
    solver.compute()
    solver.log_output()

    porosity = compute_volume_fraction(workspace, cutoff)
    eta = [porosity / solver.keff[0], porosity / solver.keff[1], porosity / solver.keff[2]]

    return eta, solver.keff, porosity, solver.T



from pumapy.utilities.workspace import Workspace


def size_check(size):
    if not (isinstance(size, tuple) and len(size) == 3):
        raise Exception("Error, invalid size, must be tuple of size 3")
    if size[0] <= 0:
        raise Exception("Error, invalid size, must be >= 0")
    if size[1] <= 0:
        raise Exception("Error, invalid size, must be >= 0")
    if size[2] <= 0:
        raise Exception("Error, invalid size, must be >= 0")


def range_inc(var, var_range, var_name):
    if var < var_range[0] or var > var_range[1]:
        raise Exception("Error, invalid " + str(var_name) + ", must be between " + str(var_range[0]) + " and " + str(
            var_range[1]) + ", inclusive")


def range_exc(var, var_range, var_name):
    if var <= var_range[0] or var >= var_range[1]:
        raise Exception("Error, invalid " + str(var_name) + ", must be between " + str(var_range[0]) + " and " + str(
            var_range[1]) + ", exclusive")


def greater_than_inc(var, lim, var_name):
    if var < lim:
        raise Exception("Error, invalid " + str(var_name) + ", must be greater than " + str(lim) + ", inclusive")


def greater_than_exc(var, lim, var_name):
    if var <= lim:
        raise Exception("Error, invalid " + str(var_name) + ", must be greater than " + str(lim) + ", exclusive")


def less_than_inc(var, lim, var_name):
    if var > lim:
        raise Exception("Error, invalid " + str(var_name) + ", must be less than " + str(lim) + ", inclusive")


def less_than_exc(var, lim, var_name):
    if var >= lim:
        raise Exception("Error, invalid " + str(var_name) + ", must be less than " + str(lim) + ", exclusive")


def check_ws_cutoff(workspace, cutoff):
    if not isinstance(workspace, Workspace):
        raise Exception("Need to pass pumapy.Workspace.")
    if workspace.matrix.shape == [1, 1, 1]:
        raise Exception("Empty workspace")
    if len(cutoff) != 2:
        raise Exception("Cutoff should be of length 2. Ex: (128,255)")
    if min(cutoff[0], cutoff[1]) < 0:
        raise Exception("Invalid cutoff. Must be positive")
    if cutoff[0] > cutoff[1]:
        raise Exception("Invalid cutoff. cutoff[0] should be <= cutoff[1]")



import unittest
import numpy as np
import pumapy as puma


class TestWorkspace(unittest.TestCase):

    def setUp(self) -> None:
        self.ws = puma.Workspace.from_shape((10, 11, 12))

    def test_WScreation(self):
        ws = puma.Workspace.from_shape((10, 11, 12))
        np.testing.assert_equal(ws.matrix, np.zeros((10, 11, 12), dtype=np.uint16))

        ws = puma.Workspace.from_shape_value((10, 11, 12), 34)
        np.testing.assert_equal(ws.matrix, np.full((10, 11, 12), 34, dtype=np.uint16))

        ws = puma.Workspace.from_array(np.full((10, 11, 12), 34))
        np.testing.assert_equal(ws.matrix, np.full((10, 11, 12), 34, dtype=np.uint16))

        ws = puma.Workspace()
        np.testing.assert_equal(ws.matrix, np.zeros((1, 1, 1), dtype=np.uint16))

        ws = puma.Workspace.from_shape((10, 11, 12), orientation=True)
        np.testing.assert_equal(ws.orientation, np.zeros((10, 11, 12, 3), dtype=np.uint16))

        ws = puma.Workspace.from_shape_vector((10, 11, 12), (1, 0, 0))
        test = np.zeros((10, 11, 12, 3), dtype=float)
        test[:] = list((1, 0, 0))
        np.testing.assert_equal(ws.orientation, test)

        ws = puma.Workspace.from_shape_value_vector((10, 11, 12), 4, (1, 0, 0))
        np.testing.assert_equal(ws.matrix, np.full((10, 11, 12), 4, dtype=np.uint16))
        test = np.zeros((10, 11, 12, 3), dtype=float)
        test[:] = list((1, 0, 0))
        np.testing.assert_equal(ws.orientation, test)

        # wrong shape
        try:
            ws = puma.Workspace.from_shape((1, 1))
        except:
            print("Correctly caught exception.")

        # wrong nparray dims
        try:
            ws = puma.Workspace.from_array(np.full((10, 12), 34))
        except:
            print("Correctly caught exception.")

        # wrong nparray type
        try:
            ws = puma.Workspace.from_array([10, 19])
        except:
            print("Correctly caught exception.")

        # wrong keyword
        try:
            ws = puma.Workspace(wrong=[10, 12])
        except:
            print("Correctly caught exception.")

        # wrong keyword
        try:
            ws = puma.Workspace(size=(20, 11, 12))
        except:
            print("Correctly caught exception.")
            return
        self.assertTrue(False)

    def test_set_voxel_length(self):
        self.ws.set_voxel_length(12)
        np.testing.assert_equal(self.ws.voxel_length, 12)

        # wrong voxel_length
        try:
            self.ws.set_voxel_length((10, 12))
        except:
            print("Correctly caught exception.")
            np.testing.assert_equal(self.ws.voxel_length, 12)
            return
        self.assertTrue(False)

    def test_set_matrix(self):
        matrix = np.random.randint(0, 255, size=self.ws.get_shape(), dtype=np.uint16)
        self.ws.set_matrix(matrix)
        np.testing.assert_equal(self.ws.matrix, matrix)

        # wrong numpy array input
        try:
            self.ws.set_matrix([1, 2, 3])
        except:
            print("Correctly caught exception.")
            np.testing.assert_equal(self.ws.matrix, matrix)

        # wrong numpy array dims
        try:
            ws2 = np.random.randint(0, 255, size=(self.ws.get_shape()[0], self.ws.get_shape()[1]), dtype=np.uint16)
            self.ws.set_matrix(ws2)
        except:
            print("Correctly caught exception.")
            np.testing.assert_equal(self.ws.matrix, matrix)
            return
        self.assertTrue(False)

    def test_copy(self):
        ws = puma.Workspace.from_shape_value((10, 11, 12), 34)
        ws2 = ws.copy()
        np.testing.assert_equal(ws.matrix, ws2.matrix)

    def test_getset(self):
        self.ws[-1, -1, -1] = 77
        np.testing.assert_equal(self.ws[-1, -1, -1], 77)

    def test_get_size(self):
        np.testing.assert_equal(self.ws.get_size(),
                                self.ws.matrix.shape[0] *
                                self.ws.matrix.shape[1] *
                                self.ws.matrix.shape[2])

    def test_len_x(self):
        np.testing.assert_equal(self.ws.len_x(), self.ws.matrix.shape[0])

    def test_len_y(self):
        np.testing.assert_equal(self.ws.len_y(), self.ws.matrix.shape[1])

    def test_ndim(self):
        np.testing.assert_equal(self.ws.ndim(), self.ws.matrix.ndim)

    def test_len_z(self):
        np.testing.assert_equal(self.ws.len_z(), self.ws.matrix.shape[2])

    def test_get_shape(self):
        np.testing.assert_equal(self.ws.get_shape(), self.ws.matrix.shape)

    def test_min(self):
        np.testing.assert_equal(self.ws.min(), self.ws.matrix.min())

    def test_max(self):
        np.testing.assert_equal(self.ws.max(), self.ws.matrix.max())

    def test_average(self):
        np.testing.assert_equal(self.ws.average(), self.ws.matrix.mean())

    def test_resize_matrix(self):
        self.ws.resize_new_matrix((12, 3, 5))
        np.testing.assert_equal(self.ws.matrix, np.zeros((12, 3, 5), dtype=np.uint16))

    def test_resize_matrixwithval(self):
        self.ws.resize_new_matrix((12, 3, 5), 4)
        np.testing.assert_equal(self.ws.matrix, np.full((12, 3, 5), 4, dtype=np.uint16))

    def test_resize_orientation(self):
        self.ws.resize_new_orientation((12, 3, 5))
        np.testing.assert_equal(self.ws.orientation, np.zeros((12, 3, 5, 3), dtype=np.uint16))

    def test_resize_orientationwithval(self):
        self.ws.resize_new_orientation((12, 3, 5), (1, 0, 0))
        np.testing.assert_equal(self.ws.orientation, np.tile([1, 0, 0], (12, 3, 5, 1)).astype(float))

    def test_set(self):
        self.ws.set(matrix_value=36, orientation_value=(1, 0, 0))
        np.testing.assert_equal(self.ws.matrix, np.full((10, 11, 12), 36, dtype=np.uint16))
        np.testing.assert_equal(self.ws.orientation, np.tile([1, 0, 0], (10, 11, 12, 1)).astype(float))

    def test_set_material_id(self):
        matrix = np.random.randint(0, 255, size=self.ws.get_shape())
        self.ws.matrix = matrix.copy()
        self.ws.set_material_id((0, 127), 0)
        self.ws.set_material_id((128, 255), 1)
        matrix[matrix <= 127] = 0
        matrix[matrix >= 128] = 1
        np.testing.assert_equal(self.ws.matrix, matrix)

    def test_binarize(self):
        matrix = np.random.randint(0, 255, size=self.ws.get_shape())
        self.ws.matrix = matrix.copy()
        self.ws.binarize(128)
        matrix[matrix < 128] = 0
        matrix[matrix >= 128] = 1
        np.testing.assert_equal(self.ws.matrix, matrix)

    def test_set_material_id_vector(self):
        test = np.zeros((10, 11, 12, 3), dtype=float)
        test[:] = list((1, 0, 0))
        self.ws = puma.Workspace.from_shape_vector((10, 11, 12), (1, 0, 0))
        self.ws.set_material_id((0, 127), 0)
        np.testing.assert_equal(self.ws.orientation, test)

    def test_binarize_vector(self):
        test = np.zeros((10, 11, 12, 3), dtype=float)
        test[:] = list((1, 0, 0))
        self.ws = puma.Workspace.from_shape_vector((10, 11, 12), (1, 0, 0))
        self.ws.binarize(128)
        np.testing.assert_equal(self.ws.orientation, test)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from pumapy.utilities.logger import print_warning
from scipy.sparse.linalg import bicgstab, spsolve, cg, gmres, minres
import inspect
import sys


class PropertySolver:

    def __init__(self, workspace, solver_type, allowed_solvers, tolerance, maxiter, display_iter):
        self.ws = workspace
        self.tolerance = tolerance
        self.maxiter = maxiter
        self.solver_type = solver_type
        self.allowed_solvers = allowed_solvers

        # First two sparse matrices need to be defined in Property child class, last two are optional
        self.Amat = None
        self.bvec = None
        self.initial_guess = None
        self.M = None

        # it returns answer in
        self.x = None

        self.del_matrices = True

        self.len_x, self.len_y, self.len_z = self.ws.matrix.shape
        self.len_xy = self.len_x * self.len_y
        self.len_xyz = self.len_xy * self.len_z

        self.callback = None
        if display_iter:
            if self.solver_type == "minres":
                self.callback = MinResSolverDisplay()
            else:
                self.callback = SolverDisplay()

    def solve(self):
        if self.solver_type not in self.allowed_solvers:
            print_warning(f"Unrecognized solver, defaulting to {self.allowed_solvers[0]}.")
            self.solver_type = self.allowed_solvers[0]

        print(f"Solving Ax=b using {self.solver_type} solver ... ", end='')

        info = 0
        if (self.solver_type == 'direct') and self.solver_type in self.allowed_solvers:
            self.x = spsolve(self.Amat, self.bvec)
        else:  # in order to use UMFPACK in spsolve, bvec needs to be a sparse matrix
            if not isinstance(self.bvec, np.ndarray):
                self.bvec = self.bvec.todense()

        # iterative solvers
        if self.solver_type == 'gmres' and self.solver_type in self.allowed_solvers:
            self.x, info = gmres(self.Amat, self.bvec, x0=self.initial_guess, M=self.M,
                                 atol=self.tolerance, maxiter=self.maxiter, callback=self.callback)

        elif self.solver_type == 'minres' and self.solver_type in self.allowed_solvers:
            self.x, info = minres(self.Amat, self.bvec, x0=self.initial_guess, M=self.M,
                                  tol=self.tolerance, maxiter=self.maxiter, callback=self.callback)

        elif self.solver_type == 'cg' and self.solver_type in self.allowed_solvers:
            self.x, info = cg(self.Amat, self.bvec, x0=self.initial_guess, M=self.M,
                              atol=self.tolerance, maxiter=self.maxiter, callback=self.callback)

        elif self.solver_type == 'bicgstab' and self.solver_type in self.allowed_solvers:
            self.x, info = bicgstab(self.Amat, self.bvec, x0=self.initial_guess, M=self.M,
                                    atol=self.tolerance, maxiter=self.maxiter, callback=self.callback)

        if info > 0:
            raise Exception("Convergence to tolerance not achieved.")
        elif info < 0:
            raise Exception("Solver illegal input or breakdown")

        if self.del_matrices:
            del self.Amat, self.bvec, self.initial_guess
        print(" ... Done")


class SolverDisplay(object):
    def __init__(self):
        self.niter = 0

    def __call__(self, rk=None):
        self.niter += 1
        frame = inspect.currentframe().f_back
        sys.stdout.write("\rIteration: {}, driving modified residual = {:0.10f} --> target = {:0.10f}"
                         .format(self.niter, frame.f_locals['resid'], frame.f_locals['atol']))


class MinResSolverDisplay(object):
    def __init__(self):
        self.niter = 0

    def __call__(self, rk=None):
        self.niter += 1
        frame = inspect.currentframe().f_back
        sys.stdout.write("\rIteration: {}, driving either residual ({:0.10f}, {:0.10f}) --> target = {:0.10f}"
                         .format(self.niter, frame.f_locals['test1'], frame.f_locals['test2'], frame.f_locals['tol']))



from pumapy.utilities.isosurface import generate_isosurface, TriMesh
from pumapy.utilities.generic_checks import check_ws_cutoff
from skimage import measure


def compute_surface_area(workspace, cutoff, flag_gaussian=False):
    """ Computation of the surface area based on isosurface

    :param workspace: domain
    :type workspace: Workspace
    :param cutoff: specify the solid phase
    :type cutoff: tuple(int, int)
    :param flag_gaussian: apply Gaussian filter before generating surface
    :type flag_gaussian: bool, optional
    :return: area, specific_area
    :rtype: tuple(float, float)
    """
    check_ws_cutoff(workspace, cutoff)

    surf_area = SurfaceArea(workspace, cutoff, flag_gaussian)

    surf_area.log_input()
    surf_area.compute()
    surf_area.log_output()

    return surf_area.area, surf_area.specific_area


class SurfaceArea:

    def __init__(self, workspace, cutoff, flag_gaussian):
        self.workspace = workspace
        self.cutoff = cutoff
        self.flag_gaussian = flag_gaussian
        self.mesh = TriMesh()
        self.area = -1.
        self.specific_area = -1.

    def compute(self):
        self.mesh = generate_isosurface(self.workspace, self.cutoff, False, self.flag_gaussian)
        self.area = self.workspace.voxel_length ** 2 * measure.mesh_surface_area(self.mesh.verts, self.mesh.faces)
        volume = self.workspace.voxel_length ** 3 * (self.workspace.len_x() - 1) * \
                 (self.workspace.len_y() - 1) * (self.workspace.len_z() - 1)
        self.specific_area = self.area / volume

    def log_input(self):
        self.workspace.log.log_section("Computing Surface Area")
        self.workspace.log.log_line("Domain Size: " + str(self.workspace.get_shape()))
        self.workspace.log.log_line("Cutoff: " + str(self.cutoff))

        if self.flag_gaussian:
            self.workspace.log.log_line("Gaussian: True")
        else:
            self.workspace.log.log_line("Gaussian: False")

        self.workspace.log.write_log()

    def log_output(self):
        self.workspace.log.log_section("Finished Surface Area Calculation")
        self.workspace.log.log_line("Surface Area: " + str(self.area))
        self.workspace.log.log_line("Specific Surface Area: " + str(self.specific_area))
